@@ -32,13 +32,14 @@
 
 #include <config.h>
 
-#ident "$Id: getdef.c 1989 2008-05-19 20:58:59Z nekral-guest $"
+#ident "$Id: getdef.c 2622 2009-04-10 22:35:26Z nekral-guest $"
 
 #include "prototypes.h"
 #include "defines.h"
 #include <stdio.h>
 #include <stdlib.h>
 #include <ctype.h>
+#include <errno.h>
 #include "getdef.h"
 /*
  * A configuration item definition.
@@ -131,7 +132,7 @@ static struct itemdef def_table[] = {
 #endif
 
 static char def_fname[] = LOGINDEFS;	/* login config defs file       */
-static int def_loaded = 0;	/* are defs already loaded?     */
+static bool def_loaded = false;		/* are defs already loaded?     */
 
 /* local function prototypes */
 static struct itemdef *def_find (const char *);
@@ -149,10 +150,12 @@ char *getdef_str (const char *item)
 {
 	struct itemdef *d;
 
-	if (!def_loaded)
+	if (!def_loaded) {
 		def_load ();
+	}
 
-	return ((d = def_find (item)) == NULL ? (char *) NULL : d->value);
+	d = def_find (item);
+	return ((NULL == d)? (char *) NULL : d->value);
 }
 
 
@@ -162,15 +165,18 @@ char *getdef_str (const char *item)
  * Return TRUE if specified item is defined as "yes", else FALSE.
  */
 
-int getdef_bool (const char *item)
+bool getdef_bool (const char *item)
 {
 	struct itemdef *d;
 
-	if (!def_loaded)
+	if (!def_loaded) {
 		def_load ();
+	}
 
-	if ((d = def_find (item)) == NULL || d->value == NULL)
-		return 0;
+	d = def_find (item);
+	if ((NULL == d) || (NULL == d->value)) {
+		return false;
+	}
 
 	return (strcasecmp (d->value, "yes") == 0);
 }
@@ -187,14 +193,27 @@ int getdef_bool (const char *item)
 int getdef_num (const char *item, int dflt)
 {
 	struct itemdef *d;
+	long val;
 
-	if (!def_loaded)
+	if (!def_loaded) {
 		def_load ();
+	}
 
-	if ((d = def_find (item)) == NULL || d->value == NULL)
+	d = def_find (item);
+	if ((NULL == d) || (NULL == d->value)) {
 		return dflt;
+	}
 
-	return (int) strtol (d->value, (char **) NULL, 0);
+	if (   (getlong (d->value, &val) == 0)
+	    || (val > INT_MAX)
+	    || (val < INT_MIN)) {
+		fprintf (stderr,
+		         _("configuration error - cannot parse %s value: '%s'"),
+		         item, d->value);
+		return dflt;
+	}
+
+	return (int) val;
 }
 
 
@@ -209,14 +228,27 @@ int getdef_num (const char *item, int dflt)
 unsigned int getdef_unum (const char *item, unsigned int dflt)
 {
 	struct itemdef *d;
+	long val;
 
-	if (!def_loaded)
+	if (!def_loaded) {
 		def_load ();
+	}
 
-	if ((d = def_find (item)) == NULL || d->value == NULL)
+	d = def_find (item);
+	if ((NULL == d) || (NULL == d->value)) {
 		return dflt;
+	}
 
-	return (unsigned int) strtoul (d->value, (char **) NULL, 0);
+	if (   (getlong (d->value, &val) == 0)
+	    || (val < 0)
+	    || (val > INT_MAX)) {
+		fprintf (stderr,
+		         _("configuration error - cannot parse %s value: '%s'"),
+		         item, d->value);
+		return dflt;
+	}
+
+	return (unsigned int) val;
 }
 
 
@@ -231,16 +263,59 @@ unsigned int getdef_unum (const char *item, unsigned int dflt)
 long getdef_long (const char *item, long dflt)
 {
 	struct itemdef *d;
+	long val;
 
-	if (!def_loaded)
+	if (!def_loaded) {
 		def_load ();
+	}
 
-	if ((d = def_find (item)) == NULL || d->value == NULL)
+	d = def_find (item);
+	if ((NULL == d) || (NULL == d->value)) {
 		return dflt;
+	}
 
-	return strtol (d->value, (char **) NULL, 0);
+	if (getlong (d->value, &val) == 0) {
+		fprintf (stderr,
+		         _("configuration error - cannot parse %s value: '%s'"),
+		         item, d->value);
+		return dflt;
+	}
+
+	return val;
 }
 
+/*
+ * getdef_ulong - get unsigned long numerical value from table of definitions
+ *
+ * Returns numeric value of specified item, else the "dflt" value if
+ * the item is not defined.  Octal (leading "0") and hex (leading "0x")
+ * values are handled.
+ */
+
+unsigned long getdef_ulong (const char *item, unsigned long dflt)
+{
+	struct itemdef *d;
+	long val;
+
+	if (!def_loaded) {
+		def_load ();
+	}
+
+	d = def_find (item);
+	if ((NULL == d) || (NULL == d->value)) {
+		return dflt;
+	}
+
+	if (getlong (d->value, &val) == 0) {
+		/* FIXME: we should have a getulong */
+		fprintf (stderr,
+		         _("configuration error - cannot parse %s value: '%s'"),
+		         item, d->value);
+		return dflt;
+	}
+
+	return val;
+}
 
 /*
  * putdef_str - override the value read from /etc/login.defs
@@ -252,28 +327,33 @@ int putdef_str (const char *name, const char *value)
 	struct itemdef *d;
 	char *cp;
 
-	if (!def_loaded)
+	if (!def_loaded) {
 		def_load ();
+	}
 
 	/*
 	 * Locate the slot to save the value.  If this parameter
 	 * is unknown then "def_find" will print an err message.
 	 */
-	if ((d = def_find (name)) == NULL)
+	d = def_find (name);
+	if (NULL == d) {
 		return -1;
+	}
 
 	/*
 	 * Save off the value.
 	 */
-	if ((cp = strdup (value)) == NULL) {
+	cp = strdup (value);
+	if (NULL == cp) {
 		fputs (_("Could not allocate space for config info.\n"),
 		       stderr);
 		SYSLOG ((LOG_ERR, "could not allocate space for config info"));
 		return -1;
 	}
 
-	if (d->value)
+	if (NULL != d->value) {
 		free (d->value);
+	}
 
 	d->value = cp;
 	return 0;
@@ -289,7 +369,6 @@ int putdef_str (const char *name, const char *value)
 
 static struct itemdef *def_find (const char *name)
 {
-	int n;
 	struct itemdef *ptr;
 
 
@@ -297,9 +376,10 @@ static struct itemdef *def_find (const char *name)
 	 * Search into the table.
 	 */
 
-	for (ptr = def_table; ptr->name; ptr++) {
-		if (!(n = strcmp (ptr->name, name)))
+	for (ptr = def_table; NULL != ptr->name; ptr++) {
+		if (strcmp (ptr->name, name) == 0) {
 			return ptr;
+		}
 	}
 
 	/*
@@ -307,9 +387,8 @@ static struct itemdef *def_find (const char *name)
 	 */
 
 	fprintf (stderr,
-		 _
-		 ("configuration error - unknown item '%s' (notify administrator)\n"),
-		 name);
+	         _("configuration error - unknown item '%s' (notify administrator)\n"),
+	         name);
 	SYSLOG ((LOG_CRIT, "unknown configuration item `%s'", name));
 	return (struct itemdef *) NULL;
 }
@@ -329,9 +408,11 @@ static void def_load (void)
 	/*
 	 * Open the configuration definitions file.
 	 */
-	if ((fp = fopen (def_fname, "r")) == NULL) {
-		SYSLOG ((LOG_CRIT, "cannot open login definitions %s [%m]",
-			 def_fname));
+	fp = fopen (def_fname, "r");
+	if (NULL == fp) {
+		int err = errno;
+		SYSLOG ((LOG_CRIT, "cannot open login definitions %s [%s]",
+		         def_fname, strerror (err)));
 		exit (1);
 	}
 
@@ -339,21 +420,23 @@ static void def_load (void)
 	 * Set the initialized flag.
 	 * (do it early to prevent recursion in putdef_str())
 	 */
-	++def_loaded;
+	def_loaded = true;
 
 	/*
 	 * Go through all of the lines in the file.
 	 */
-	while (fgets (buf, sizeof (buf), fp) != NULL) {
+	while (fgets (buf, (int) sizeof (buf), fp) != NULL) {
 
 		/*
 		 * Trim trailing whitespace.
 		 */
-		for (i = strlen (buf) - 1; i >= 0; --i) {
-			if (!isspace (buf[i]))
+		for (i = (int) strlen (buf) - 1; i >= 0; --i) {
+			if (!isspace (buf[i])) {
 				break;
+			}
 		}
-		buf[++i] = '\0';
+		i++;
+		buf[i] = '\0';
 
 		/*
 		 * Break the line into two fields.
@@ -376,9 +459,10 @@ static void def_load (void)
 		putdef_str (name, value);
 	}
 
-	if (ferror (fp)) {
-		SYSLOG ((LOG_CRIT, "cannot read login definitions %s [%m]",
-			 def_fname));
+	if (ferror (fp) != 0) {
+		int err = errno;
+		SYSLOG ((LOG_CRIT, "cannot read login definitions %s [%s]",
+		         def_fname, strerror (err)));
 		exit (1);
 	}
 
@@ -396,17 +480,21 @@ int main (int argc, char **argv)
 	def_load ();
 
 	for (i = 0; i < NUMDEFS; ++i) {
-		if ((d = def_find (def_table[i].name)) == NULL)
+		d = def_find (def_table[i].name);
+		if (NULL == d) {
 			printf ("error - lookup '%s' failed\n",
-				def_table[i].name);
-		else
+			        def_table[i].name);
+		} else {
 			printf ("%4d %-24s %s\n", i + 1, d->name, d->value);
+		}
 	}
 	for (i = 1; i < argc; i++) {
-		if ((cp = getdef_str (argv[1])) != NULL)
+		cp = getdef_str (argv[1]);
+		if (NULL != cp) {
 			printf ("%s `%s'\n", argv[1], cp);
-		else
+		} else {
 			printf ("%s not found\n", argv[1]);
+		}
 	}
 	exit (0);
 }
