@@ -1,5 +1,8 @@
 /*
- * Copyright 1989 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1989 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1996 - 1998, Marek Michałkiewicz
+ * Copyright (c) 2001 - 2006, Tomasz Kłoczko
+ * Copyright (c) 2008       , Nicolas François
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -10,21 +13,21 @@
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. Neither the name of Julianne F. Haugh nor the names of its contributors
- *    may be used to endorse or promote products derived from this software
- *    without specific prior written permission.
+ * 3. The name of the copyright holders or contributors may not be used to
+ *    endorse or promote products derived from this software without
+ *    specific prior written permission.
  *
- * THIS SOFTWARE IS PROVIDED BY JULIE HAUGH AND CONTRIBUTORS ``AS IS'' AND
- * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED.  IN NO EVENT SHALL JULIE HAUGH OR CONTRIBUTORS BE LIABLE
- * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
- * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
- * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
+ * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
 #include <config.h>
@@ -35,10 +38,11 @@
 #include <errno.h>
 #include "prototypes.h"
 #include "defines.h"
+#include "exitcodes.h"
 #include <pwd.h>
 #include <grp.h>
 
-#ident "$Id: age.c,v 1.12 2005/08/31 17:24:57 kloczek Exp $"
+#ident "$Id: age.c 2650 2009-04-11 18:37:08Z nekral-guest $"
 
 #ifndef PASSWD_PROGRAM
 #define PASSWD_PROGRAM "/bin/passwd"
@@ -52,28 +56,30 @@
 int expire (const struct passwd *pw, const struct spwd *sp)
 {
 	int status;
-	int child;
-	int pid;
+	pid_t child;
+	pid_t pid;
 
-	if (!sp)
-		sp = pwd_to_spwd (pw);
+	if (NULL == sp) {
+		return 0;
+	}
 
 	/*
 	 * See if the user's password has expired, and if so
 	 * force them to change their password.
 	 */
 
-	switch (status = isexpired (pw, sp)) {
+	status = isexpired (pw, sp);
+	switch (status) {
 	case 0:
 		return 0;
 	case 1:
-		printf (_("Your password has expired."));
+		(void) fputs (_("Your password has expired."), stdout);
 		break;
 	case 2:
-		printf (_("Your password is inactive."));
+		(void) fputs (_("Your password is inactive."), stdout);
 		break;
 	case 3:
-		printf (_("Your login has expired."));
+		(void) fputs (_("Your login has expired."), stdout);
 		break;
 	}
 
@@ -84,12 +90,12 @@ int expire (const struct passwd *pw, const struct spwd *sp)
 	 * change that password.
 	 */
 
-	if (status > 1 || sp->sp_max < sp->sp_min) {
-		puts (_("  Contact the system administrator.\n"));
+	if ((status > 1) || (sp->sp_max < sp->sp_min)) {
+		(void) puts (_("  Contact the system administrator."));
 		exit (1);
 	}
-	puts (_("  Choose a new password.\n"));
-	fflush (stdout);
+	(void) puts (_("  Choose a new password."));
+	(void) fflush (stdout);
 
 	/*
 	 * Close all the files so that unauthorized access won't
@@ -111,7 +117,8 @@ int expire (const struct passwd *pw, const struct spwd *sp)
 	 * change their password before being able to use the account.
 	 */
 
-	if ((pid = fork ()) == 0) {
+	pid = fork ();
+	if (0 == pid) {
 		int err;
 
 		/*
@@ -119,21 +126,29 @@ int expire (const struct passwd *pw, const struct spwd *sp)
 		 * passwd to work just like it would had they executed
 		 * it from the command line while logged in.
 		 */
-		if (setup_uid_gid (pw, 0))
+#if defined(HAVE_INITGROUPS) && ! defined(USE_PAM)
+		if (setup_uid_gid (pw, false) != 0)
+#else
+		if (setup_uid_gid (pw) != 0)
+#endif
+		{
 			_exit (126);
+		}
 
 		execl (PASSWD_PROGRAM, PASSWD_PROGRAM, pw->pw_name, (char *) 0);
 		err = errno;
 		perror ("Can't execute " PASSWD_PROGRAM);
-		_exit ((err == ENOENT) ? 127 : 126);
-	} else if (pid == -1) {
+		_exit ((ENOENT == err) ? E_CMD_NOTFOUND : E_CMD_NOEXEC);
+	} else if ((pid_t) -1 == pid) {
 		perror ("fork");
 		exit (1);
 	}
-	while ((child = wait (&status)) != pid && child != -1);
 
-	if (child == pid && status == 0)
+	while (((child = wait (&status)) != pid) && (child != (pid_t)-1));
+
+	if ((child == pid) && (0 == status)) {
 		return 1;
+	}
 
 	exit (1);
  /*NOTREACHED*/}
@@ -145,30 +160,42 @@ int expire (const struct passwd *pw, const struct spwd *sp)
  *	to expire and warns the user of the pending password expiration.
  */
 
-void agecheck (const struct passwd *pw, const struct spwd *sp)
+void agecheck (const struct spwd *sp)
 {
-	long now = time ((long *) 0) / SCALE;
+	long now = (long) time ((time_t *) 0) / SCALE;
 	long remain;
 
-	if (!sp)
-		sp = pwd_to_spwd (pw);
+	if (NULL == sp) {
+		return;
+	}
 
 	/*
 	 * The last, max, and warn fields must be supported or the
 	 * warning period cannot be calculated.
 	 */
 
-	if (sp->sp_lstchg == -1 || sp->sp_max == -1 || sp->sp_warn == -1)
+	if (   (-1 == sp->sp_lstchg)
+	    || (-1 == sp->sp_max)
+	    || (-1 == sp->sp_warn)) {
 		return;
-	if ((remain = (sp->sp_lstchg + sp->sp_max) - now) <= sp->sp_warn) {
+	}
+
+	if (0 == sp->sp_lstchg) {
+		(void) puts (_("You must change your password."));
+		return;
+	}
+
+	remain = sp->sp_lstchg + sp->sp_max - now;
+	if (remain <= sp->sp_warn) {
 		remain /= DAY / SCALE;
-		if (remain > 1)
-			printf (_
-				("Your password will expire in %ld days.\n"),
-				remain);
-		else if (remain == 1)
-			printf (_("Your password will expire tomorrow.\n"));
-		else if (remain == 0)
-			printf (_("Your password will expire today.\n"));
+		if (remain > 1) {
+			(void) printf (_("Your password will expire in %ld days.\n"),
+			               remain);
+		} else if (1 == remain) {
+			(void) puts (_("Your password will expire tomorrow."));
+		} else if (remain == 0) {
+			(void) puts (_("Your password will expire today."));
+		}
 	}
 }
+
