@@ -32,7 +32,7 @@
 
 #include <config.h>
 
-#ident "$Id: copydir.c 2007 2008-05-24 15:35:15Z nekral-guest $"
+#ident "$Id: copydir.c 2643 2009-04-11 15:34:10Z nekral-guest $"
 
 #include <sys/stat.h>
 #include <sys/types.h>
@@ -43,7 +43,6 @@
 #include "defines.h"
 #ifdef WITH_SELINUX
 #include <selinux/selinux.h>
-static int selinux_enabled = -1;
 #endif
 static const char *src_orig;
 static const char *dst_orig;
@@ -60,20 +59,20 @@ static struct link_name *links;
 static int copy_entry (const char *src, const char *dst,
                        long int uid, long int gid);
 static int copy_dir (const char *src, const char *dst,
-                     const struct stat *statp, const struct timeval mt[2],
+                     const struct stat *statp, const struct timeval mt[],
                      long int uid, long int gid);
 #ifdef	S_IFLNK
 static int copy_symlink (const char *src, const char *dst,
-                         const struct stat *statp, const struct timeval mt[2],
+                         const struct stat *statp, const struct timeval mt[],
                          long int uid, long int gid);
 #endif
 static int copy_hardlink (const char *src, const char *dst,
                           struct link_name *lp);
 static int copy_special (const char *dst,
-                         const struct stat *statp, const struct timeval mt[2],
+                         const struct stat *statp, const struct timeval mt[],
                          long int uid, long int gid);
 static int copy_file (const char *src, const char *dst,
-                      const struct stat *statp, const struct timeval mt[2],
+                      const struct stat *statp, const struct timeval mt[],
                       long int uid, long int gid);
 
 #ifdef WITH_SELINUX
@@ -84,14 +83,21 @@ static int copy_file (const char *src, const char *dst,
  *	selinux_file_context () should be called before any creation of file,
  *	symlink, directory, ...
  *
+ *	Callers may have to Reset SELinux to create files with default
+ *	contexts:
+ *		setfscreatecon (NULL);
  */
-static int selinux_file_context (const char *dst_name)
+int selinux_file_context (const char *dst_name)
 {
+	static bool selinux_checked = false;
+	static bool selinux_enabled;
 	security_context_t scontext = NULL;
 
-	if (selinux_enabled < 0) {
+	if (!selinux_checked) {
 		selinux_enabled = is_selinux_enabled () > 0;
+		selinux_checked = true;
 	}
+
 	if (selinux_enabled) {
 		/* Get the default security context for this file */
 		if (matchpathcon (dst_name, 0, &scontext) < 0) {
@@ -146,10 +152,10 @@ static void remove_link (struct link_name *ln)
 static struct link_name *check_link (const char *name, const struct stat *sb)
 {
 	struct link_name *lp;
-	int src_len;
-	int dst_len;
-	int name_len;
-	int len;
+	size_t src_len;
+	size_t dst_len;
+	size_t name_len;
+	size_t len;
 
 	for (lp = links; lp; lp = lp->ln_next) {
 		if ((lp->ln_dev == sb->st_dev) && (lp->ln_ino == sb->st_ino)) {
@@ -169,7 +175,7 @@ static struct link_name *check_link (const char *name, const struct stat *sb)
 	lp->ln_ino = sb->st_ino;
 	lp->ln_count = sb->st_nlink;
 	len = name_len - src_len + dst_len + 1;
-	lp->ln_name = xmalloc (len);
+	lp->ln_name = (char *) xmalloc (len);
 	snprintf (lp->ln_name, len, "%s%s", dst_orig, name + src_len);
 	lp->ln_next = links;
 	links = lp;
@@ -189,7 +195,7 @@ int copy_tree (const char *src_root, const char *dst_root,
 	char src_name[1024];
 	char dst_name[1024];
 	int err = 0;
-	int set_orig = 0;
+	bool set_orig = false;
 	struct DIRECT *ent;
 	DIR *dir;
 
@@ -216,10 +222,10 @@ int copy_tree (const char *src_root, const char *dst_root,
 		return -1;
 	}
 
-	if (src_orig == 0) {
+	if (src_orig == NULL) {
 		src_orig = src_root;
 		dst_orig = dst_root;
-		set_orig++;
+		set_orig = true;
 	}
 	while ((0 == err) && (ent = readdir (dir)) != NULL) {
 		/*
@@ -253,9 +259,15 @@ int copy_tree (const char *src_root, const char *dst_root,
 	(void) closedir (dir);
 
 	if (set_orig) {
-		src_orig = 0;
-		dst_orig = 0;
+		src_orig = NULL;
+		dst_orig = NULL;
 	}
+
+#ifdef WITH_SELINUX
+	/* Reset SELinux to create files with default contexts */
+	setfscreatecon (NULL);
+#endif
+
 	return err;
 }
 
@@ -285,16 +297,28 @@ static int copy_entry (const char *src, const char *dst,
 	if (LSTAT (src, &sb) == -1) {
 		/* If we cannot stat the file, do not care. */
 	} else {
-#if  defined(_BSD_SOURCE) || defined(_SVID_SOURCE)
+#ifdef HAVE_STRUCT_STAT_ST_ATIM
 		mt[0].tv_sec  = sb.st_atim.tv_sec;
 		mt[0].tv_usec = sb.st_atim.tv_nsec / 1000;
+#else
+		mt[0].tv_sec  = sb.st_atime;
+#ifdef HAVE_STRUCT_STAT_ST_ATIMENSEC
+		mt[0].tv_usec = sb.st_atimensec / 1000;
+#else
+		mt[0].tv_usec = 0;
+#endif
+#endif
+
+#ifdef HAVE_STRUCT_STAT_ST_MTIM
 		mt[1].tv_sec  = sb.st_mtim.tv_sec;
 		mt[1].tv_usec = sb.st_mtim.tv_nsec / 1000;
 #else
-		mt[0].tv_sec  = sb.st_atime;
-		mt[0].tv_usec = sb.st_atimensec / 1000;
 		mt[1].tv_sec  = sb.st_mtime;
+#ifdef HAVE_STRUCT_STAT_ST_MTIMENSEC
 		mt[1].tv_usec = sb.st_mtimensec / 1000;
+#else
+		mt[1].tv_usec = 0;
+#endif
 #endif
 
 		if (S_ISDIR (sb.st_mode)) {
@@ -353,7 +377,7 @@ static int copy_entry (const char *src, const char *dst,
  *	Return 0 on success, -1 on error.
  */
 static int copy_dir (const char *src, const char *dst,
-                     const struct stat *statp, const struct timeval mt[2],
+                     const struct stat *statp, const struct timeval mt[],
                      long int uid, long int gid)
 {
 	int err = 0;
@@ -391,7 +415,7 @@ static int copy_dir (const char *src, const char *dst,
  *	Return 0 on success, -1 on error.
  */
 static int copy_symlink (const char *src, const char *dst,
-                         const struct stat *statp, const struct timeval mt[2],
+                         const struct stat *statp, const struct timeval mt[],
                          long int uid, long int gid)
 {
 	char oldlink[1024];
@@ -428,12 +452,14 @@ static int copy_symlink (const char *src, const char *dst,
 		return -1;
 	}
 
+#ifdef HAVE_LUTIMES
 	/* 2007-10-18: We don't care about
 	 *  exit status of lutimes because
 	 *  it returns ENOSYS on many system
 	 *  - not implemented
 	 */
 	lutimes (dst, mt);
+#endif
 
 	return err;
 }
@@ -479,7 +505,7 @@ static int copy_hardlink (const char *src, const char *dst,
  *	Return 0 on success, -1 on error.
  */
 static int copy_special (const char *dst,
-                         const struct stat *statp, const struct timeval mt[2],
+                         const struct stat *statp, const struct timeval mt[],
                          long int uid, long int gid)
 {
 	int err = 0;
@@ -511,14 +537,14 @@ static int copy_special (const char *dst,
  *	Return 0 on success, -1 on error.
  */
 static int copy_file (const char *src, const char *dst,
-                      const struct stat *statp, const struct timeval mt[2],
+                      const struct stat *statp, const struct timeval mt[],
                       long int uid, long int gid)
 {
 	int err = 0;
 	int ifd;
 	int ofd;
 	char buf[1024];
-	int cnt;
+	ssize_t cnt;
 
 	ifd = open (src, O_RDONLY);
 	if (ifd < 0) {
@@ -538,21 +564,29 @@ static int copy_file (const char *src, const char *dst,
 	}
 
 	while ((cnt = read (ifd, buf, sizeof buf)) > 0) {
-		if (write (ofd, buf, cnt) != cnt) {
+		if (write (ofd, buf, (size_t)cnt) != cnt) {
 			return -1;
 		}
 	}
 
 	(void) close (ifd);
 
+#ifdef HAVE_FUTIMES
 	if (futimes (ofd, mt) != 0) {
 		return -1;
 	}
+#endif
 
 	if (close (ofd) != 0) {
 		return -1;
 	}
 
+#ifndef HAVE_FUTIMES
+	if (utimes(dst, mt) != 0) {
+		return -1;
+	}
+#endif
+
 	return err;
 }
 
