@@ -2,6 +2,7 @@
  * Copyright (c) 1989 - 1994, Julianne Frances Haugh
  * Copyright (c) 1996 - 1997, Marek Michałkiewicz
  * Copyright (c) 2001 - 2005, Tomasz Kłoczko
+ * Copyright (c) 2008       , Nicolas François
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -42,7 +43,7 @@
 #include <pwd.h>
 #include <time.h>
 
-#ident "$Id: isexpired.c 1980 2008-04-27 00:40:09Z nekral-guest $"
+#ident "$Id: isexpired.c 2592 2009-04-05 22:02:00Z nekral-guest $"
 
 
 /*
@@ -50,15 +51,23 @@
  *
  *	isexpired calculates the expiration date based on the
  *	password expiration criteria.
+ *
+ * Return value:
+ *	0: The password is still valid
+ *	1: The password has expired, it must be changed
+ *	2: The password has expired since a long time and the account is
+ *	   now disabled. (password cannot be changed)
+ *	3: The account has expired
  */
- /*ARGSUSED*/ int isexpired (const struct passwd *pw, const struct spwd *sp)
+int isexpired (const struct passwd *pw, const struct spwd *sp)
 {
 	long now;
 
-	now = time ((time_t *) 0) / SCALE;
+	now = (long) time ((time_t *) 0) / SCALE;
 
-	if (!sp)
-		sp = pwd_to_spwd (pw);
+	if (NULL == sp) {
+		return 0;
+	}
 
 	/*
 	 * Quick and easy - there is an expired account field
@@ -66,8 +75,9 @@
 	 * one first since it is worse.
 	 */
 
-	if (sp->sp_expire > 0 && now >= sp->sp_expire)
+	if ((sp->sp_expire > 0) && (now >= sp->sp_expire)) {
 		return 3;
+	}
 
 	/*
 	 * Last changed date 1970-01-01 (not very likely) means that
@@ -77,12 +87,17 @@
 	 * if /etc/shadow doesn't exist, getspnam() still succeeds and
 	 * returns sp_lstchg==0 (must change password) instead of -1!
 	 */
-	if (sp->sp_lstchg == 0 && !strcmp (pw->pw_passwd, SHADOW_PASSWD_STRING))
+	if (   (0 == sp->sp_lstchg)
+	    && (strcmp (pw->pw_passwd, SHADOW_PASSWD_STRING) == 0)) {
 		return 1;
+	}
 
-	if (sp->sp_lstchg > 0 && sp->sp_max >= 0 && sp->sp_inact >= 0 &&
-	    now >= sp->sp_lstchg + sp->sp_max + sp->sp_inact)
+	if (   (sp->sp_lstchg > 0)
+	    && (sp->sp_max >= 0)
+	    && (sp->sp_inact >= 0)
+	    && (now >= (sp->sp_lstchg + sp->sp_max + sp->sp_inact))) {
 		return 2;
+	}
 
 	/*
 	 * The last and max fields must be present for an account
@@ -90,9 +105,11 @@
 	 * is considered to be infinite.
 	 */
 
-	if (sp->sp_lstchg == -1 ||
-	    sp->sp_max == -1 || sp->sp_max >= (10000L * DAY / SCALE))
+	if (   (-1 == sp->sp_lstchg)
+	    || (-1 == sp->sp_max)
+	    || (sp->sp_max >= (10000L * DAY / SCALE))) {
 		return 0;
+	}
 
 	/*
 	 * Calculate today's day and the day on which the password
@@ -100,7 +117,9 @@
 	 * the password has expired.
 	 */
 
-	if (now >= sp->sp_lstchg + sp->sp_max)
+	if (now >= (sp->sp_lstchg + sp->sp_max)) {
 		return 1;
+	}
 	return 0;
 }
+
