@@ -44,7 +44,7 @@
 #include <fcntl.h>
 #include <stdio.h>
 
-#ident "$Id: utmp.c 2450 2008-11-22 23:54:54Z nekral-guest $"
+#ident "$Id: utmp.c 2451 2008-11-22 23:56:11Z nekral-guest $"
 
 #if HAVE_UTMPX_H
 struct utmpx utxent;
@@ -59,7 +59,7 @@ struct utmp utent;
 /*
  * is_my_tty -- determine if "tty" is the same TTY stdin is using
  */
-static int is_my_tty (const char *tty)
+static bool is_my_tty (const char *tty)
 {
 	char full_tty[200];
 	struct stat by_name, by_fd;
@@ -69,14 +69,15 @@ static int is_my_tty (const char *tty)
 		tty = full_tty;
 	}
 
-	if ((stat (tty, &by_name) != 0) || (fstat (STDIN_FILENO, &by_fd) != 0)) {
-		return 0;
+	if (   (stat (tty, &by_name) != 0)
+	    || (fstat (STDIN_FILENO, &by_fd) != 0)) {
+		return false;
 	}
 
 	if (by_name.st_rdev != by_fd.st_rdev) {
-		return 0;
+		return false;
 	} else {
-		return 1;
+		return true;
 	}
 }
 
@@ -96,7 +97,7 @@ static int is_my_tty (const char *tty)
 
 #if defined(__linux__)		/* XXX */
 
-void checkutmp (int picky)
+void checkutmp (bool picky)
 {
 	char *line;
 	struct utmp *ut;
@@ -105,33 +106,40 @@ void checkutmp (int picky)
 	setutent ();
 
 	/* First, try to find a valid utmp entry for this process.  */
-	while ((ut = getutent ()))
-		if (ut->ut_pid == pid && ut->ut_line[0] && ut->ut_id[0] &&
-		    (ut->ut_type == LOGIN_PROCESS
-		     || ut->ut_type == USER_PROCESS) &&
-		    is_my_tty (ut->ut_line))
+	while ((ut = getutent ()) != NULL) {
+		if (   (ut->ut_pid == pid)
+		    && ('\0' != ut->ut_line[0])
+		    && ('\0' != ut->ut_id[0])
+		    && (   (LOGIN_PROCESS == ut->ut_type)
+		        || (USER_PROCESS  == ut->ut_type))
+		    /* A process may have failed to close an entry
+		     * Check if this entry refers to this tty */
+		    && is_my_tty (ut->ut_line)) {
 			break;
+		}
+	}
 
-	/* If there is one, just use it, otherwise create a new one.  */
-	if (ut) {
+	/* If there is one, just use it, otherwise create a new one. */
+	if (NULL != ut) {
 		utent = *ut;
 	} else {
 		if (picky) {
-			puts (NO_UTENT);
-			exit (1);
+			(void) puts (NO_UTENT);
+			exit (EXIT_FAILURE);
 		}
 		line = ttyname (0);
-		if (!line) {
-			puts (NO_TTY);
-			exit (1);
+		if (NULL == line) {
+			(void) puts (NO_TTY);
+			exit (EXIT_FAILURE);
 		}
-		if (strncmp (line, "/dev/", 5) == 0)
+		if (strncmp (line, "/dev/", 5) == 0) {
 			line += 5;
+		}
 		memset ((void *) &utent, 0, sizeof utent);
 		utent.ut_type = LOGIN_PROCESS;
 		utent.ut_pid = pid;
 		strncpy (utent.ut_line, line, sizeof utent.ut_line);
-		/* XXX - assumes /dev/tty?? */
+		/* XXX - assumes /dev/tty?? or /dev/pts/?? */
 		strncpy (utent.ut_id, utent.ut_line + 3, sizeof utent.ut_id);
 		strcpy (utent.ut_user, "LOGIN");
 		utent.ut_time = time (NULL);
@@ -140,7 +148,7 @@ void checkutmp (int picky)
 
 #elif defined(LOGIN_PROCESS)
 
-void checkutmp (int picky)
+void checkutmp (bool picky)
 {
 	char *line;
 	struct utmp *ut;
@@ -157,28 +165,34 @@ void checkutmp (int picky)
 
 	if (picky) {
 #if HAVE_UTMPX_H
-		while ((utx = getutxent ()))
-			if (utx->ut_pid == pid)
+		while ((utx = getutxent ()) != NULL) {
+			if (utx->ut_pid == pid) {
 				break;
+			}
+		}
 
-		if (utx)
+		if (NULL != utx) {
 			utxent = *utx;
+		}
 #endif
-		while ((ut = getutent ()))
-			if (ut->ut_pid == pid)
+		while ((ut = getutent ()) != NULL) {
+			if (ut->ut_pid == pid) {
 				break;
+			}
+		}
 
-		if (ut)
+		if (NULL != ut) {
 			utent = *ut;
+		}
 
 #if HAVE_UTMPX_H
 		endutxent ();
 #endif
 		endutent ();
 
-		if (!ut) {
-			puts (NO_UTENT);
-			exit (1);
+		if (NULL == ut) {
+			(void) puts (NO_UTENT);
+			exit (EXIT_FAILURE);
 		}
 #ifndef	UNIXPC
 
@@ -188,32 +202,42 @@ void checkutmp (int picky)
 		 * the structure.  The UNIX/PC is broken in this regard
 		 * and needs help ...
 		 */
+		/* XXX: The ut_line may not match with the current tty.
+		 *      ut_line will be set by setutmp anyway, but ut_id
+		 *      will not be set, and the wrong utmp entry may be
+		 *      updated. -- nekral */
 
 		if (utent.ut_line[0] == '\0')
 #endif				/* !UNIXPC */
 		{
-			if (!(line = ttyname (0))) {
-				puts (NO_TTY);
-				exit (1);
+			line = ttyname (0);
+			if (NULL == line) {
+				(void) puts (NO_TTY);
+				exit (EXIT_FAILURE);
 			}
-			if (strncmp (line, "/dev/", 5) == 0)
+			if (strncmp (line, "/dev/", 5) == 0) {
 				line += 5;
+			}
 			strncpy (utent.ut_line, line, sizeof utent.ut_line);
 #if HAVE_UTMPX_H
 			strncpy (utxent.ut_line, line, sizeof utxent.ut_line);
 #endif
 		}
 	} else {
-		if (!(line = ttyname (0))) {
-			puts (NO_TTY);
-			exit (1);
+		line = ttyname (0);
+		if (NULL == line) {
+			(void) puts (NO_TTY);
+			exit (EXIT_FAILURE);
 		}
-		if (strncmp (line, "/dev/", 5) == 0)
+		if (strncmp (line, "/dev/", 5) == 0) {
 			line += 5;
+		}
 
 		strncpy (utent.ut_line, line, sizeof utent.ut_line);
-		if ((ut = getutline (&utent)))
+		ut = getutline (&utent);
+		if (NULL != ut) {
 			strncpy (utent.ut_id, ut->ut_id, sizeof ut->ut_id);
+		}
 
 		strcpy (utent.ut_user, "LOGIN");
 		utent.ut_pid = getpid ();
@@ -221,15 +245,17 @@ void checkutmp (int picky)
 		utent.ut_time = time (NULL);
 #if HAVE_UTMPX_H
 		strncpy (utxent.ut_line, line, sizeof utxent.ut_line);
-		if ((utx = getutxline (&utxent)))
+		utx = getutxline (&utxent);
+		if (NULL != utx) {
 			strncpy (utxent.ut_id, utx->ut_id, sizeof utxent.ut_id);
+		}
 
 		strcpy (utxent.ut_user, "LOGIN");
 		utxent.ut_pid = utent.ut_pid;
 		utxent.ut_type = utent.ut_type;
-		if (sizeof (utxent.ut_tv) == sizeof (struct timeval))
+		if (sizeof (utxent.ut_tv) == sizeof (struct timeval)) {
 			gettimeofday ((struct timeval *) &utxent.ut_tv, NULL);
-		else {
+		} else {
 			struct timeval tv;
 
 			gettimeofday (&tv, NULL);
@@ -287,34 +313,42 @@ static void updwtmpx (const char *filename, const struct utmpx *utx)
 
 #if defined(__linux__)		/* XXX */
 
-void setutmp (const char *name, const char *line, const char *host)
+int setutmp (const char *name, const char unused(*line), const char unused(*host))
 {
+	int err = 0;
 	utent.ut_type = USER_PROCESS;
 	strncpy (utent.ut_user, name, sizeof utent.ut_user);
 	utent.ut_time = time (NULL);
 	/* other fields already filled in by checkutmp above */
 	setutent ();
-	pututline (&utent);
+	if (pututline (&utent) == NULL) {
+		err = 1;
+	}
 	endutent ();
 	updwtmp (_WTMP_FILE, &utent);
+
+	return err;
 }
 
 #elif HAVE_UTMPX_H
 
-void setutmp (const char *name, const char *line, const char *host)
+int setutmp (const char *name, const char *line, const char *host)
 {
 	struct utmp *utmp, utline;
 	struct utmpx *utmpx, utxline;
 	pid_t pid = getpid ();
-	int found_utmpx = 0, found_utmp = 0;
+	bool found_utmpx = false;
+	bool found_utmp = false;
+	int err = 0;
 
 	/*
 	 * The canonical device name doesn't include "/dev/"; skip it
 	 * if it is already there.
 	 */
 
-	if (strncmp (line, "/dev/", 5) == 0)
+	if (strncmp (line, "/dev/", 5) == 0) {
 		line += 5;
+	}
 
 	/*
 	 * Update utmpx.  We create an empty entry in case there is
@@ -324,15 +358,15 @@ void setutmp (const char *name, const char *line, const char *host)
 	setutxent ();
 	setutent ();
 
-	while (utmpx = getutxent ()) {
+	while ((utmpx = getutxent ()) != NULL) {
 		if (utmpx->ut_pid == pid) {
-			found_utmpx = 1;
+			found_utmpx = true;
 			break;
 		}
 	}
-	while (utmp = getutent ()) {
+	while ((utmp = getutent ()) != NULL) {
 		if (utmp->ut_pid == pid) {
-			found_utmp = 1;
+			found_utmp = true;
 			break;
 		}
 	}
@@ -379,9 +413,9 @@ void setutmp (const char *name, const char *line, const char *host)
 
 	utline.ut_type = utxline.ut_type = USER_PROCESS;
 
-	if (sizeof (utxline.ut_tv) == sizeof (struct timeval))
+	if (sizeof (utxline.ut_tv) == sizeof (struct timeval)) {
 		gettimeofday ((struct timeval *) &utxline.ut_tv, NULL);
-	else {
+	} else {
 		struct timeval tv;
 
 		gettimeofday (&tv, NULL);
@@ -390,16 +424,21 @@ void setutmp (const char *name, const char *line, const char *host)
 	}
 	utline.ut_time = utxline.ut_tv.tv_sec;
 
-	strncpy (utxline.ut_host, host ? host : "", sizeof utxline.ut_host);
+	strncpy (utxline.ut_host, (NULL != host) ? host : "",
+	         sizeof utxline.ut_host);
 
-	pututxline (&utxline);
-	pututline (&utline);
+	if (   (pututxline (&utxline) == NULL)
+	    || (pututline (&utline) == NULL)) {
+		err = 1;
+	}
 
 	updwtmpx (_WTMP_FILE "x", &utxline);
 	updwtmp (_WTMP_FILE, &utline);
 
 	utxent = utxline;
 	utent = utline;
+
+	return err;
 }
 
 #endif
