@@ -3,7 +3,7 @@
  * Copyright (c) 1996 - 2000, Marek Michałkiewicz
  * Copyright (c) 2001       , Michał Moskal
  * Copyright (c) 2001 - 2006, Tomasz Kłoczko
- * Copyright (c) 2007 - 2008, Nicolas François
+ * Copyright (c) 2007 - 2009, Nicolas François
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -33,7 +33,7 @@
 
 #include <config.h>
 
-#ident "$Id: grpck.c 1980 2008-04-27 00:40:09Z nekral-guest $"
+#ident "$Id: grpck.c 2583 2009-03-22 12:32:40Z nekral-guest $"
 
 #include <fcntl.h>
 #include <grp.h>
@@ -64,41 +64,73 @@
 /*
  * Global variables
  */
-static char *Prog;
+char *Prog;
+
 static const char *grp_file = GROUP_FILE;
-static int use_system_grp_file = 1;
+static bool use_system_grp_file = true;
 
 #ifdef	SHADOWGRP
 static const char *sgr_file = SGROUP_FILE;
-static int use_system_sgr_file = 1;
-static int is_shadow = 0;
+static bool use_system_sgr_file = true;
+static bool is_shadow = false;
+static bool sgr_locked = false;
 #endif
+static bool gr_locked = false;
 /* Options */
-static int read_only = 0;
-static int sort_mode = 0;
+static bool read_only = false;
+static bool sort_mode = false;
 
 /* local function prototypes */
+static void fail_exit (int status);
 static void usage (void);
 static void delete_member (char **, const char *);
 static void process_flags (int argc, char **argv);
 static void open_files (void);
-static void close_files (int changed);
+static void close_files (bool changed);
 static int check_members (const char *groupname,
                           char **members,
                           const char *fmt_info,
                           const char *fmt_prompt,
                           const char *fmt_syslog,
                           int *errors);
-static void check_grp_file (int *errors, int *changed);
+static void check_grp_file (int *errors, bool *changed);
 #ifdef SHADOWGRP
 static void compare_members_lists (const char *groupname,
                                    char **members,
                                    char **other_members,
                                    const char *file,
                                    const char *other_file);
-static void check_sgr_file (int *errors, int *changed);
+static void check_sgr_file (int *errors, bool *changed);
 #endif
 
+/*
+ * fail_exit - exit with an error code after unlocking files
+ */
+static void fail_exit (int status)
+{
+	if (gr_locked) {
+		if (gr_unlock () == 0) {
+			fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, gr_dbname ());
+			SYSLOG ((LOG_ERR, "failed to unlock %s", gr_dbname ()));
+			/* continue */
+		}
+	}
+
+#ifdef	SHADOWGRP
+	if (sgr_locked) {
+		if (sgr_unlock () == 0) {
+			fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, sgr_dbname ());
+			SYSLOG ((LOG_ERR, "failed to unlock %s", sgr_dbname ()));
+			/* continue */
+		}
+	}
+#endif
+
+	closelog ();
+
+	exit (status);
+}
+
 /*
  * usage - print syntax message and exit
  */
@@ -114,6 +146,8 @@ static void usage (void)
 
 /*
  * delete_member - delete an entry in a list of members
+ *
+ * It only deletes the first entry with the given name.
  */
 static void delete_member (char **list, const char *member)
 {
@@ -150,10 +184,10 @@ static void process_flags (int argc, char **argv)
 			/* quiet - ignored for now */
 			break;
 		case 'r':
-			read_only = 1;
+			read_only = true;
 			break;
 		case 's':
-			sort_mode = 1;
+			sort_mode = true;
 			break;
 		default:
 			usage ();
@@ -183,15 +217,15 @@ static void process_flags (int argc, char **argv)
 	 */
 	if (optind != argc) {
 		grp_file = argv[optind];
-		gr_name (grp_file);
-		use_system_grp_file = 0;
+		gr_setdbname (grp_file);
+		use_system_grp_file = false;
 	}
 #ifdef	SHADOWGRP
 	if ((optind + 2) == argc) {
 		sgr_file = argv[optind + 1];
-		sgr_name (sgr_file);
-		is_shadow = 1;
-		use_system_sgr_file = 0;
+		sgr_setdbname (sgr_file);
+		is_shadow = true;
+		use_system_sgr_file = false;
 	} else if (optind == argc) {
 		is_shadow = sgr_file_present ();
 	}
@@ -211,23 +245,21 @@ static void open_files (void)
 	 */
 	if (!read_only) {
 		if (gr_lock () == 0) {
-			fprintf (stderr, _("%s: cannot lock file %s\n"),
+			fprintf (stderr,
+			         _("%s: cannot lock %s; try again later.\n"),
 			         Prog, grp_file);
-			if (use_system_grp_file) {
-				SYSLOG ((LOG_WARN, "cannot lock %s", grp_file));
-			}
-			closelog ();
-			exit (E_CANT_LOCK);
+			fail_exit (E_CANT_LOCK);
 		}
+		gr_locked = true;
 #ifdef	SHADOWGRP
-		if (is_shadow && (sgr_lock () == 0)) {
-			fprintf (stderr, _("%s: cannot lock file %s\n"),
-			         Prog, sgr_file);
-			if (use_system_sgr_file) {
-				SYSLOG ((LOG_WARN, "cannot lock %s", sgr_file));
+		if (is_shadow) {
+			if (sgr_lock () == 0) {
+				fprintf (stderr,
+				         _("%s: cannot lock %s; try again later.\n"),
+				         Prog, sgr_file);
+				fail_exit (E_CANT_LOCK);
 			}
-			closelog ();
-			exit (E_CANT_LOCK);
+			sgr_locked = true;
 		}
 #endif
 	}
@@ -237,23 +269,21 @@ static void open_files (void)
 	 * O_RDWR otherwise.
 	 */
 	if (gr_open (read_only ? O_RDONLY : O_RDWR) == 0) {
-		fprintf (stderr, _("%s: cannot open file %s\n"), Prog,
+		fprintf (stderr, _("%s: cannot open %s\n"), Prog,
 		         grp_file);
 		if (use_system_grp_file) {
 			SYSLOG ((LOG_WARN, "cannot open %s", grp_file));
 		}
-		closelog ();
-		exit (E_CANT_OPEN);
+		fail_exit (E_CANT_OPEN);
 	}
 #ifdef	SHADOWGRP
 	if (is_shadow && (sgr_open (read_only ? O_RDONLY : O_RDWR) == 0)) {
-		fprintf (stderr, _("%s: cannot open file %s\n"), Prog,
+		fprintf (stderr, _("%s: cannot open %s\n"), Prog,
 		         sgr_file);
 		if (use_system_sgr_file) {
 			SYSLOG ((LOG_WARN, "cannot open %s", sgr_file));
 		}
-		closelog ();
-		exit (E_CANT_OPEN);
+		fail_exit (E_CANT_OPEN);
 	}
 #endif
 }
@@ -265,7 +295,7 @@ static void open_files (void)
  *	changes are committed in the databases. The databases are
  *	unlocked anyway.
  */
-static void close_files (int changed)
+static void close_files (bool changed)
 {
 	/*
 	 * All done. If there were no change we can just abandon any
@@ -273,15 +303,15 @@ static void close_files (int changed)
 	 */
 	if (changed) {
 		if (gr_close () == 0) {
-			fprintf (stderr, _("%s: cannot update file %s\n"),
+			fprintf (stderr, _("%s: failure while writing changes to %s\n"),
 			         Prog, grp_file);
-			exit (E_CANT_UPDATE);
+			fail_exit (E_CANT_UPDATE);
 		}
 #ifdef	SHADOWGRP
 		if (is_shadow && (sgr_close () == 0)) {
-			fprintf (stderr, _("%s: cannot update file %s\n"),
+			fprintf (stderr, _("%s: failure while writing changes to %s\n"),
 			         Prog, sgr_file);
-			exit (E_CANT_UPDATE);
+			fail_exit (E_CANT_UPDATE);
 		}
 #endif
 	}
@@ -290,11 +320,23 @@ static void close_files (int changed)
 	 * Don't be anti-social - unlock the files when you're done.
 	 */
 #ifdef	SHADOWGRP
-	if (is_shadow) {
-		sgr_unlock ();
+	if (sgr_locked) {
+		if (sgr_unlock () == 0) {
+			fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, sgr_dbname ());
+			SYSLOG ((LOG_ERR, "failed to unlock %s", sgr_dbname ()));
+			/* continue */
+		}
+		sgr_locked = false;
 	}
 #endif
-	(void) gr_unlock ();
+	if (gr_locked) {
+		if (gr_unlock () == 0) {
+			fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, gr_dbname ());
+			SYSLOG ((LOG_ERR, "failed to unlock %s", gr_dbname ()));
+			/* continue */
+		}
+		gr_locked = false;
+	}
 }
 
 /*
@@ -349,6 +391,9 @@ static int check_members (const char *groupname,
 		SYSLOG ((LOG_INFO, fmt_syslog, members[i], groupname));
 		members_changed = 1;
 		delete_member (members, members[i]);
+
+		/* Rewind in case of removal */
+		i--;
 	}
 
 	return members_changed;
@@ -391,7 +436,7 @@ static void compare_members_lists (const char *groupname,
 /*
  * check_grp_file - check the content of the group file
  */
-static void check_grp_file (int *errors, int *changed)
+static void check_grp_file (int *errors, bool *changed)
 {
 	struct commonio_entry *gre, *tgre;
 	struct group *grp;
@@ -440,9 +485,9 @@ static void check_grp_file (int *errors, int *changed)
 			 * to try out the next list element.
 			 */
 		      delete_gr:
-			SYSLOG ((LOG_INFO, "delete group line `%s'",
+			SYSLOG ((LOG_INFO, "delete group line '%s'",
 			         gre->line));
-			*changed = 1;
+			*changed = true;
 
 			__gr_del_entry (gre);
 			continue;
@@ -497,11 +542,19 @@ static void check_grp_file (int *errors, int *changed)
 		/*
 		 * Check for invalid group names.  --marekm
 		 */
-		if (!check_group_name (grp->gr_name)) {
+		if (!is_valid_group_name (grp->gr_name)) {
 			*errors += 1;
 			printf (_("invalid group name '%s'\n"), grp->gr_name);
 		}
 
+		/*
+		 * Check for invalid group ID.
+		 */
+		if (grp->gr_gid == (gid_t)-1) {
+			printf (_("invalid group ID '%lu'\n"), (long unsigned int)grp->gr_gid);
+			*errors += 1;
+		}
+
 		/*
 		 * Workaround for a NYS libc 5.3.12 bug on RedHat 4.2 -
 		 * groups with no members are returned as groups with one
@@ -516,10 +569,10 @@ static void check_grp_file (int *errors, int *changed)
 		if (check_members (grp->gr_name, grp->gr_mem,
 		                   _("group %s: no user %s\n"),
 		                   _("delete member '%s'? "),
-		                   "delete member `%s' from group `%s'",
+		                   "delete member '%s' from group '%s'",
 		                   errors) == 1) {
-			*changed = 1;
-			gre->changed = 1;
+			*changed = true;
+			gre->changed = true;
 			__gr_set_changed ();
 		}
 
@@ -531,10 +584,9 @@ static void check_grp_file (int *errors, int *changed)
 		if (is_shadow) {
 			sgr = (struct sgrp *) sgr_locate (grp->gr_name);
 			if (sgr == NULL) {
-				printf (_
-				        ("no matching group file entry in %s\n"),
+				printf (_("no matching group file entry in %s\n"),
 				        sgr_file);
-				printf (_("add group '%s' in %s ?"),
+				printf (_("add group '%s' in %s? "),
 				        grp->gr_name, sgr_file);
 				*errors += 1;
 				if (yes_or_no (read_only)) {
@@ -547,26 +599,24 @@ static void check_grp_file (int *errors, int *changed)
 					sg.sg_adm = &empty;
 					sg.sg_mem = grp->gr_mem;
 					SYSLOG ((LOG_INFO,
-					         "add group `%s' to `%s'",
+					         "add group '%s' to '%s'",
 					         grp->gr_name, sgr_file));
-					*changed = 1;
+					*changed = true;
 
 					if (sgr_update (&sg) == 0) {
 						fprintf (stderr,
-						         _
-						         ("%s: can't update shadow entry for %s\n"),
-						         Prog, sg.sg_name);
-						exit (E_CANT_UPDATE);
+						         _("%s: failed to prepare the new %s entry '%s'\n"),
+						         Prog, sgr_dbname (), sg.sg_name);
+						fail_exit (E_CANT_UPDATE);
 					}
 					/* remove password from /etc/group */
 					gr = *grp;
 					gr.gr_passwd = SHADOW_PASSWD_STRING;	/* XXX warning: const */
 					if (gr_update (&gr) == 0) {
 						fprintf (stderr,
-						         _
-						         ("%s: can't update entry for group %s\n"),
-						         Prog, gr.gr_name);
-						exit (E_CANT_UPDATE);
+						         _("%s: failed to prepare the new %s entry '%s'\n"),
+						         Prog, gr_dbname (), gr.gr_name);
+						fail_exit (E_CANT_UPDATE);
 					}
 				}
 			} else {
@@ -588,7 +638,7 @@ static void check_grp_file (int *errors, int *changed)
 /*
  * check_sgr_file - check the content of the shadowed group file (gshadow)
  */
-static void check_sgr_file (int *errors, int *changed)
+static void check_sgr_file (int *errors, bool *changed)
 {
 	struct group *grp;
 	struct commonio_entry *sge, *tsge;
@@ -628,9 +678,9 @@ static void check_sgr_file (int *errors, int *changed)
 			 * of the loop to try out the next list element.
 			 */
 		      delete_sg:
-			SYSLOG ((LOG_INFO, "delete shadow line `%s'",
+			SYSLOG ((LOG_INFO, "delete shadow line '%s'",
 			         sge->line));
-			*changed = 1;
+			*changed = true;
 
 			__sgr_del_entry (sge);
 			continue;
@@ -710,10 +760,10 @@ static void check_sgr_file (int *errors, int *changed)
 		if (check_members (sgr->sg_name, sgr->sg_adm,
 		                   _("shadow group %s: no administrative user %s\n"),
 		                   _("delete administrative member '%s'? "),
-		                   "delete admin `%s' from shadow group `%s'",
+		                   "delete admin '%s' from shadow group '%s'",
 		                   errors) == 1) {
-			*changed = 1;
-			sge->changed = 1;
+			*changed = true;
+			sge->changed = true;
 			__sgr_set_changed ();
 		}
 
@@ -723,10 +773,10 @@ static void check_sgr_file (int *errors, int *changed)
 		if (check_members (sgr->sg_name, sgr->sg_mem,
 		                   _("shadow group %s: no user %s\n"),
 		                   _("delete member '%s'? "),
-		                   "delete member `%s' from shadow group `%s'",
+		                   "delete member '%s' from shadow group '%s'",
 		                   errors) == 1) {
-			*changed = 1;
-			sge->changed = 1;
+			*changed = true;
+			sge->changed = true;
 			__sgr_set_changed ();
 		}
 	}
@@ -739,16 +789,16 @@ static void check_sgr_file (int *errors, int *changed)
 int main (int argc, char **argv)
 {
 	int errors = 0;
-	int changed = 0;
+	bool changed = false;
 
 	/*
 	 * Get my name so that I can use it to report errors.
 	 */
 	Prog = Basename (argv[0]);
 
-	setlocale (LC_ALL, "");
-	bindtextdomain (PACKAGE, LOCALEDIR);
-	textdomain (PACKAGE);
+	(void) setlocale (LC_ALL, "");
+	(void) bindtextdomain (PACKAGE, LOCALEDIR);
+	(void) textdomain (PACKAGE);
 
 	OPENLOG ("grpck");
 
@@ -763,7 +813,7 @@ int main (int argc, char **argv)
 		if (is_shadow) {
 			sgr_sort ();
 		}
-		changed = 1;
+		changed = true;
 #endif
 	} else {
 		check_grp_file (&errors, &changed);
@@ -782,12 +832,12 @@ int main (int argc, char **argv)
 	/*
 	 * Tell the user what we did and exit.
 	 */
-	if (errors != 0) {
+	if (0 != errors) {
 		printf (changed ?
 		        _("%s: the files have been updated\n") :
 		        _("%s: no changes\n"), Prog);
 	}
 
-	exit (errors ? E_BAD_ENTRY : E_OKAY);
+	exit ((0 != errors) ? E_BAD_ENTRY : E_OKAY);
 }
 
