@@ -2,7 +2,7 @@
  * Copyright (c) 1990 - 1993, Julianne Frances Haugh
  * Copyright (c) 1996 - 2000, Marek Michałkiewicz
  * Copyright (c) 2000 - 2006, Tomasz Kłoczko
- * Copyright (c) 2007 - 2008, Nicolas François
+ * Copyright (c) 2007 - 2009, Nicolas François
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -40,7 +40,7 @@
 
 #include <config.h>
 
-#ident "$Id: newusers.c 1980 2008-04-27 00:40:09Z nekral-guest $"
+#ident "$Id: newusers.c 2659 2009-04-11 23:30:44Z nekral-guest $"
 
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -50,9 +50,13 @@
 #include <fcntl.h>
 #include <getopt.h>
 #include <ctype.h>
+#include <errno.h>
+#include <string.h>
+#ifdef ACCT_TOOLS_SETUID
 #ifdef USE_PAM
 #include "pam_defs.h"
 #endif				/* USE_PAM */
+#endif				/* ACCT_TOOLS_SETUID */
 #include "prototypes.h"
 #include "defines.h"
 #include "getdef.h"
@@ -62,35 +66,37 @@
 #include "sgroupio.h"
 #include "shadowio.h"
 #include "chkname.h"
+
 /*
  * Global variables
  */
-static char *Prog;
-static int cflg = 0;
-static int rflg = 0;	/* create a system account */
-static int sflg = 0;
+char *Prog;
+
+static bool cflg = false;
+static bool rflg = false;	/* create a system account */
+#ifdef USE_SHA_CRYPT
+static bool sflg = false;
+#endif
 
 static char *crypt_method = NULL;
+#ifdef USE_SHA_CRYPT
 static long sha_rounds = 5000;
+#endif
 
-static int is_shadow;
+static bool is_shadow;
 #ifdef SHADOWGRP
-static int is_shadow_grp;
-static int gshadow_locked = 0;
-#endif
-static int passwd_locked = 0;
-static int group_locked = 0;
-static int shadow_locked = 0;
-
-#ifdef USE_PAM
-static pam_handle_t *pamh = NULL;
+static bool is_shadow_grp;
+static bool sgr_locked = false;
 #endif
+static bool pw_locked = false;
+static bool gr_locked = false;
+static bool spw_locked = false;
 
 /* local function prototypes */
 static void usage (void);
 static void fail_exit (int);
 static int add_group (const char *, const char *, gid_t *, gid_t);
-static int get_uid (const char *, uid_t *);
+static int get_user_id (const char *, uid_t *);
 static int add_user (const char *, uid_t, gid_t);
 static void update_passwd (struct passwd *, const char *);
 static int add_passwd (struct passwd *, const char *);
@@ -128,18 +134,34 @@ static void usage (void)
  */
 static void fail_exit (int code)
 {
-	if (shadow_locked) {
-		spw_unlock ();
+	if (spw_locked) {
+		if (spw_unlock () == 0) {
+			fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, spw_dbname ());
+			SYSLOG ((LOG_ERR, "failed to unlock %s", spw_dbname ()));
+			/* continue */
+		}
 	}
-	if (passwd_locked) {
-		pw_unlock ();
+	if (pw_locked) {
+		if (pw_unlock () == 0) {
+			fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, pw_dbname ());
+			SYSLOG ((LOG_ERR, "failed to unlock %s", pw_dbname ()));
+			/* continue */
+		}
 	}
-	if (group_locked) {
-		gr_unlock ();
+	if (gr_locked) {
+		if (gr_unlock () == 0) {
+			fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, gr_dbname ());
+			SYSLOG ((LOG_ERR, "failed to unlock %s", gr_dbname ()));
+			/* continue */
+		}
 	}
 #ifdef	SHADOWGRP
-	if (gshadow_locked) {
-		sgr_unlock ();
+	if (sgr_locked) {
+		if (sgr_unlock () == 0) {
+			fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, sgr_dbname ());
+			SYSLOG ((LOG_ERR, "failed to unlock %s", sgr_dbname ()));
+			/* continue */
+		}
 	}
 #endif
 
@@ -178,22 +200,32 @@ static int add_group (const char *name, const char *gid, gid_t *ngid, uid_t uid)
 		 * The GID is a number, which means either this is a brand
 		 * new group, or an existing group.
 		 */
-		char *endptr;
-		long int i = strtoul (gid, &endptr, 10);
-		if ((*endptr != '\0') && (errno != ERANGE)) {
+
+		if (get_gid (gid, &grent.gr_gid) == 0) {
 			fprintf (stderr,
-			         _("%s: group ID `%s' is not valid\n"),
+			         _("%s: invalid group ID '%s'\n"),
 			         Prog, gid);
 			return -1;
 		}
-		if (   (getgrgid (i) != NULL)
-		    || (gr_locate_gid (i) != NULL)) {
+
+		/* Look in both the system database (getgrgid) and in the
+		 * internal database (gr_locate_gid), which may contain
+		 * uncommitted changes */
+		if (   (getgrgid ((gid_t) grent.gr_gid) != NULL)
+		    || (gr_locate_gid ((gid_t) grent.gr_gid) != NULL)) {
 			/* The user will use this ID for her
 			 * primary group */
-			*ngid = i;
+			*ngid = (gid_t) grent.gr_gid;
 			return 0;
 		}
-		grent.gr_gid = i;
+
+		/* Do not create groups with GID == (gid_t)-1 */
+		if (grent.gr_gid == (gid_t)-1) {
+			fprintf (stderr,
+			         _("%s: invalid group ID '%s'\n"),
+			         Prog, gid);
+			return -1;
+		}
 	} else {
 		/* The gid parameter can be "" or a name which is not
 		 * already the name of an existing group.
@@ -211,12 +243,13 @@ static int add_group (const char *name, const char *gid, gid_t *ngid, uid_t uid)
 		grent.gr_name = xstrdup (gid);
 	} else {
 		grent.gr_name = xstrdup (name);
+/* FIXME: check if the group exist */
 	}
 
 	/* Check if this is a valid group name */
-	if (check_group_name (grent.gr_name) == 0) {
+	if (!is_valid_group_name (grent.gr_name)) {
 		fprintf (stderr,
-		         _("%s: invalid group name `%s'\n"),
+		         _("%s: invalid group name '%s'\n"),
 		         Prog, grent.gr_name);
 		free (grent.gr_name);
 		return -1;
@@ -234,7 +267,7 @@ static int add_group (const char *name, const char *gid, gid_t *ngid, uid_t uid)
 
 		if (NULL != sg) {
 			fprintf (stderr,
-			         _("%s: group %s is a shadow group, but does not exist in /etc/group\n"),
+			         _("%s: group '%s' is a shadow group, but does not exist in /etc/group\n"),
 			         Prog, grent.gr_name);
 			return -1;
 		}
@@ -256,9 +289,6 @@ static int add_group (const char *name, const char *gid, gid_t *ngid, uid_t uid)
 		sgrent.sg_mem = members;
 
 		if (sgr_update (&sgrent) == 0) {
-			fprintf (stderr,
-			         _("%s: group %s created, failure during the creation of the corresponding gshadow group\n"),
-			         Prog, grent.gr_name);
 			return -1;
 		}
 	}
@@ -267,25 +297,22 @@ static int add_group (const char *name, const char *gid, gid_t *ngid, uid_t uid)
 	return 0;
 }
 
-static int get_uid (const char *uid, uid_t *nuid) {
-	const struct passwd *pwd = NULL;
+static int get_user_id (const char *uid, uid_t *nuid) {
 
 	/*
 	 * The first guess for the UID is either the numerical UID that the
 	 * caller provided, or the next available UID.
 	 */
 	if (isdigit (uid[0])) {
-		char *endptr;
-		long int i = strtoul (uid, &endptr, 10);
-		if ((*endptr != '\0') && (errno != ERANGE)) {
+		if ((get_uid (uid, nuid) == 0) || (*nuid == (uid_t)-1)) {
 			fprintf (stderr,
-			         _("%s: user ID `%s' is not valid\n"),
+			         _("%s: invalid user ID '%s'\n"),
 			         Prog, uid);
 			return -1;
 		}
-		*nuid = i;
 	} else {
 		if ('\0' != uid[0]) {
+			const struct passwd *pwd;
 			/* local, no need for xgetpwnam */
 			pwd = getpwnam (uid);
 			if (NULL == pwd) {
@@ -296,7 +323,7 @@ static int get_uid (const char *uid, uid_t *nuid) {
 				*nuid = pwd->pw_uid;
 			} else {
 				fprintf (stderr,
-				         _("%s: user `%s' does not exist\n"),
+				         _("%s: user '%s' does not exist\n"),
 				         Prog, uid);
 				return -1;
 			}
@@ -318,9 +345,9 @@ static int add_user (const char *name, uid_t uid, gid_t gid)
 	struct passwd pwent;
 
 	/* Check if this is a valid user name */
-	if (check_user_name (name) == 0) {
+	if (!is_valid_user_name (name)) {
 		fprintf (stderr,
-		         _("%s: invalid user name `%s'\n"),
+		         _("%s: invalid user name '%s'\n"),
 		         Prog, name);
 		return -1;
 	}
@@ -338,16 +365,18 @@ static int add_user (const char *name, uid_t uid, gid_t gid)
 	pwent.pw_dir = "";	/* XXX warning: const */
 	pwent.pw_shell = "";	/* XXX warning: const */
 
-	return !pw_update (&pwent);
+	return (pw_update (&pwent) == 0);
 }
 
 static void update_passwd (struct passwd *pwd, const char *password)
 {
 	void *crypt_arg = NULL;
 	if (crypt_method != NULL) {
+#ifdef USE_SHA_CRYPT
 		if (sflg) {
 			crypt_arg = &sha_rounds;
 		}
+#endif
 	}
 
 	if ((crypt_method != NULL) && (0 == strcmp(crypt_method, "NONE"))) {
@@ -368,9 +397,11 @@ static int add_passwd (struct passwd *pwd, const char *password)
 	struct spwd spent;
 	void *crypt_arg = NULL;
 	if (crypt_method != NULL) {
+#ifdef USE_SHA_CRYPT
 		if (sflg) {
 			crypt_arg = &sha_rounds;
 		}
+#endif
 	}
 
 	/*
@@ -390,7 +421,7 @@ static int add_passwd (struct passwd *pwd, const char *password)
 	sp = spw_locate (pwd->pw_name);
 	if (NULL != sp) {
 		spent = *sp;
-		if (   (crypt_method != NULL)
+		if (   (NULL != crypt_method)
 		    && (0 == strcmp(crypt_method, "NONE"))) {
 			spent.sp_pwdp = (char *)password;
 		} else {
@@ -398,7 +429,7 @@ static int add_passwd (struct passwd *pwd, const char *password)
 			                                    crypt_arg);
 			spent.sp_pwdp = pw_encrypt (password, salt);
 		}
-		return !spw_update (&spent);
+		return (spw_update (&spent) == 0);
 	}
 
 	/*
@@ -423,16 +454,20 @@ static int add_passwd (struct passwd *pwd, const char *password)
 		const char *salt = crypt_make_salt (crypt_method, crypt_arg);
 		spent.sp_pwdp = pw_encrypt (password, salt);
 	}
-	spent.sp_lstchg = time ((time_t *) 0) / SCALE;
-	spent.sp_min = getdef_num ("PASS_MIN_DAYS", 0);
+	spent.sp_lstchg = (long) time ((time_t *) 0) / SCALE;
+	if (0 == spent.sp_lstchg) {
+		/* Better disable aging than requiring a password change */
+		spent.sp_lstchg = -1;
+	}
+	spent.sp_min    = getdef_num ("PASS_MIN_DAYS", 0);
 	/* 10000 is infinity this week */
-	spent.sp_max = getdef_num ("PASS_MAX_DAYS", 10000);
-	spent.sp_warn = getdef_num ("PASS_WARN_AGE", -1);
-	spent.sp_inact = -1;
+	spent.sp_max    = getdef_num ("PASS_MAX_DAYS", 10000);
+	spent.sp_warn   = getdef_num ("PASS_WARN_AGE", -1);
+	spent.sp_inact  = -1;
 	spent.sp_expire = -1;
-	spent.sp_flag = -1;
+	spent.sp_flag   = SHADOW_SP_FLAG_UNSET;
 
-	return !spw_update (&spent);
+	return (spw_update (&spent) == 0);
 }
 
 /*
@@ -450,28 +485,32 @@ static void process_flags (int argc, char **argv)
 #ifdef USE_SHA_CRYPT
 		{"sha-rounds", required_argument, NULL, 's'},
 #endif
+		{"system", no_argument, NULL, 'r'},
 		{NULL, 0, NULL, '\0'}
 	};
 
 	while ((c = getopt_long (argc, argv,
 #ifdef USE_SHA_CRYPT
-	                     "c:hs:",
+	                         "c:hrs:",
 #else
-	                     "c:h",
+	                         "c:hr",
 #endif
 	                     long_options, &option_index)) != -1) {
 		switch (c) {
 		case 'c':
-			cflg = 1;
+			cflg = true;
 			crypt_method = optarg;
 			break;
 		case 'h':
 			usage ();
 			break;
+		case 'r':
+			rflg = true;
+			break;
 #ifdef USE_SHA_CRYPT
 		case 's':
-			sflg = 1;
-			if (!getlong(optarg, &sha_rounds)) {
+			sflg = true;
+			if (getlong(optarg, &sha_rounds) == 0) {
 				fprintf (stderr,
 				         _("%s: invalid numeric argument '%s'\n"),
 				         Prog, optarg);
@@ -489,7 +528,7 @@ static void process_flags (int argc, char **argv)
 	}
 
 	if (argv[optind] != NULL) {
-		if (!freopen (argv[optind], "r", stdin)) {
+		if (freopen (argv[optind], "r", stdin) == NULL) {
 			char buf[BUFSIZ];
 			snprintf (buf, sizeof buf, "%s: %s", Prog, argv[1]);
 			perror (buf);
@@ -508,12 +547,14 @@ static void process_flags (int argc, char **argv)
  */
 static void check_flags (void)
 {
+#ifdef USE_SHA_CRYPT
 	if (sflg && !cflg) {
 		fprintf (stderr,
-		         _("%s: %s flag is ONLY allowed with the %s flag\n"),
+		         _("%s: %s flag is only allowed with the %s flag\n"),
 		         Prog, "-s", "-c");
 		usage ();
 	}
+#endif
 
 	if (cflg) {
 		if (   (0 != strcmp (crypt_method, "DES"))
@@ -544,38 +585,39 @@ static void check_flags (void)
  */
 static void check_perms (void)
 {
+#ifdef ACCT_TOOLS_SETUID
 #ifdef USE_PAM
-	int retval = PAM_SUCCESS;
+	pam_handle_t *pamh = NULL;
+	int retval;
 	struct passwd *pampw;
 
 	pampw = getpwuid (getuid ()); /* local, no need for xgetpwuid */
-	if (pampw == NULL) {
-		retval = PAM_USER_UNKNOWN;
+	if (NULL == pampw) {
+		fprintf (stderr,
+		         _("%s: Cannot determine your user name.\n"),
+		         Prog);
+		fail_exit (1);
 	}
 
-	if (retval == PAM_SUCCESS) {
-		retval = pam_start ("newusers", pampw->pw_name, &conv, &pamh);
-	}
+	retval = pam_start ("newusers", pampw->pw_name, &conv, &pamh);
 
-	if (retval == PAM_SUCCESS) {
+	if (PAM_SUCCESS == retval) {
 		retval = pam_authenticate (pamh, 0);
-		if (retval != PAM_SUCCESS) {
-			pam_end (pamh, retval);
-		}
 	}
 
-	if (retval == PAM_SUCCESS) {
+	if (PAM_SUCCESS == retval) {
 		retval = pam_acct_mgmt (pamh, 0);
-		if (retval != PAM_SUCCESS) {
-			pam_end (pamh, retval);
-		}
 	}
 
-	if (retval != PAM_SUCCESS) {
+	if (NULL != pamh) {
+		(void) pam_end (pamh, retval);
+	}
+	if (PAM_SUCCESS != retval) {
 		fprintf (stderr, _("%s: PAM authentication failed\n"), Prog);
 		fail_exit (1);
 	}
 #endif				/* USE_PAM */
+#endif				/* ACCT_TOOLS_SETUID */
 }
 
 /*
@@ -589,39 +631,59 @@ static void open_files (void)
 	 * modified, or new entries added. The password file is the key - if
 	 * it gets locked, assume the others can be locked right away.
 	 */
-	if (!pw_lock ()) {
-		fprintf (stderr, _("%s: can't lock /etc/passwd.\n"), Prog);
+	if (pw_lock () == 0) {
+		fprintf (stderr,
+		         _("%s: cannot lock %s; try again later.\n"),
+		         Prog, pw_dbname ());
 		fail_exit (1);
 	}
-	passwd_locked++;
-	if (is_shadow && !spw_lock ()) {
-		fprintf (stderr, _("%s: can't lock /etc/shadow.\n"), Prog);
+	pw_locked = true;
+	if (is_shadow) {
+		if (spw_lock () == 0) {
+			fprintf (stderr,
+			         _("%s: cannot lock %s; try again later.\n"),
+			         Prog, spw_dbname ());
+			fail_exit (1);
+		}
+		spw_locked = true;
+	}
+	if (gr_lock () == 0) {
+		fprintf (stderr,
+		         _("%s: cannot lock %s; try again later.\n"),
+		         Prog, gr_dbname ());
 		fail_exit (1);
 	}
-	shadow_locked++;
-	if (!gr_lock ()) {
-		fprintf (stderr, _("%s: can't lock /etc/group.\n"), Prog);
-		fail_exit (1);
-	}
-	group_locked++;
+	gr_locked = true;
 #ifdef SHADOWGRP
-	if (is_shadow_grp && !sgr_lock ()) {
-		fprintf (stderr, _("%s: can't lock /etc/gshadow.\n"), Prog);
-		fail_exit (1);
+	if (is_shadow_grp) {
+		if (sgr_lock () == 0) {
+			fprintf (stderr,
+			         _("%s: cannot lock %s; try again later.\n"),
+			         Prog, sgr_dbname ());
+			fail_exit (1);
+		}
+		sgr_locked = true;
 	}
-	gshadow_locked++;
 #endif
 
-	if (   (!pw_open (O_RDWR))
-	    || (is_shadow && !spw_open (O_RDWR))
-	    || !gr_open (O_RDWR)
-#ifdef SHADOWGRP
-	    || (is_shadow_grp && !sgr_open(O_RDWR))
-#endif
-	   ) {
-		fprintf (stderr, _("%s: can't open files\n"), Prog);
+	if (pw_open (O_RDWR) == 0) {
+		fprintf (stderr, _("%s: cannot open %s\n"), Prog, pw_dbname ());
 		fail_exit (1);
 	}
+	if (is_shadow && (spw_open (O_RDWR) == 0)) {
+		fprintf (stderr, _("%s: cannot open %s\n"), Prog, spw_dbname ());
+		fail_exit (1);
+	}
+	if (gr_open (O_RDWR) == 0) {
+		fprintf (stderr, _("%s: cannot open %s\n"), Prog, gr_dbname ());
+		fail_exit (1);
+	}
+#ifdef SHADOWGRP
+	if (is_shadow_grp && (sgr_open (O_RDWR) == 0)) {
+		fprintf (stderr, _("%s: cannot open %s\n"), Prog, sgr_dbname ());
+		fail_exit (1);
+	}
+#endif
 }
 
 /*
@@ -629,30 +691,71 @@ static void open_files (void)
  */
 static void close_files (void)
 {
-	if (   (!pw_close ())
-	    || (is_shadow && !spw_close ())
-	    || !gr_close ()
-#ifdef SHADOWGRP
-	    || (is_shadow_grp && !sgr_close())
-#endif
-	   ) {
-		fprintf (stderr, _("%s: error updating files\n"), Prog);
+	if (pw_close () == 0) {
+		fprintf (stderr, _("%s: failure while writing changes to %s\n"), Prog, pw_dbname ());
+		SYSLOG ((LOG_ERR, "failure while writing changes to %s", pw_dbname ()));
 		fail_exit (1);
 	}
+	if (pw_unlock () == 0) {
+		fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, pw_dbname ());
+		SYSLOG ((LOG_ERR, "failed to unlock %s", pw_dbname ()));
+		/* continue */
+	}
+	pw_locked = false;
+
+	if (is_shadow) {
+		if (spw_close () == 0) {
+			fprintf (stderr,
+			         _("%s: failure while writing changes to %s\n"),
+			         Prog, spw_dbname ());
+			SYSLOG ((LOG_ERR, "failure while writing changes to %s", spw_dbname ()));
+			fail_exit (1);
+		}
+		if (spw_unlock () == 0) {
+			fprintf (stderr,
+			         _("%s: failed to unlock %s\n"),
+			         Prog, spw_dbname ());
+			SYSLOG ((LOG_ERR, "failed to unlock %s", spw_dbname ()));
+			/* continue */
+		}
+		spw_locked = false;
+	}
+
+	if (gr_close () == 0) {
+		fprintf (stderr,
+		         _("%s: failure while writing changes to %s\n"),
+		         Prog, gr_dbname ());
+		SYSLOG ((LOG_ERR, "failure while writing changes to %s", gr_dbname ()));
+		fail_exit (1);
+	}
+	if (gr_unlock () == 0) {
+		fprintf (stderr,
+		         _("%s: failed to unlock %s\n"),
+		         Prog, gr_dbname ());
+		SYSLOG ((LOG_ERR, "failed to unlock %s", gr_dbname ()));
+		/* continue */
+	}
+	gr_locked = false;
+
 #ifdef SHADOWGRP
 	if (is_shadow_grp) {
-		(void) sgr_unlock();
-		gshadow_locked--;
+		if (sgr_close () == 0) {
+			fprintf (stderr,
+			         _("%s: failure while writing changes to %s\n"),
+			         Prog, sgr_dbname ());
+			SYSLOG ((LOG_ERR, "failure while writing changes to %s", sgr_dbname ()));
+			fail_exit (1);
+		}
+		if (sgr_unlock () == 0) {
+			fprintf (stderr,
+			         _("%s: failed to unlock %s\n"),
+			         Prog, sgr_dbname ());
+			SYSLOG ((LOG_ERR, "failed to unlock %s", sgr_dbname ()));
+			/* continue */
+		}
+		sgr_locked = false;
 	}
 #endif
-	(void) gr_unlock ();
-	group_locked--;
-	if (is_shadow) {
-		(void) spw_unlock ();
-		shadow_locked--;
-	}
-	(void) pw_unlock ();
-	passwd_locked--;
 }
 
 int main (int argc, char **argv)
@@ -670,9 +773,11 @@ int main (int argc, char **argv)
 
 	Prog = Basename (argv[0]);
 
-	setlocale (LC_ALL, "");
-	bindtextdomain (PACKAGE, LOCALEDIR);
-	textdomain (PACKAGE);
+	(void) setlocale (LC_ALL, "");
+	(void) bindtextdomain (PACKAGE, LOCALEDIR);
+	(void) textdomain (PACKAGE);
+
+	OPENLOG ("newusers");
 
 	process_flags (argc, argv);
 
@@ -695,7 +800,7 @@ int main (int argc, char **argv)
 	 * over 100 is allocated. The pw_gid field will be updated with that
 	 * value.
 	 */
-	while (fgets (buf, sizeof buf, stdin) != (char *) 0) {
+	while (fgets (buf, (int) sizeof buf, stdin) != (char *) 0) {
 		line++;
 		cp = strrchr (buf, '\n');
 		if (NULL != cp) {
@@ -716,7 +821,8 @@ int main (int argc, char **argv)
 			fields[nfields] = cp;
 			cp = strchr (cp, ':');
 			if (NULL != cp) {
-				*cp++ = '\0';
+				*cp = '\0';
+				cp++;
 			} else {
 				break;
 			}
@@ -740,7 +846,7 @@ int main (int argc, char **argv)
 		}
 
 		if (   (NULL == pw)
-		    && (get_uid (fields[2], &uid) != 0)) {
+		    && (get_user_id (fields[2], &uid) != 0)) {
 			fprintf (stderr,
 			         _("%s: line %d: can't create user\n"),
 			         Prog, line);
@@ -792,8 +898,8 @@ int main (int argc, char **argv)
 		pw = pw_locate (fields[0]);
 		if (NULL == pw) {
 			fprintf (stderr,
-			         _("%s: line %d: cannot find user %s\n"),
-			         Prog, line, fields[0]);
+			         _("%s: line %d: user '%s' does not exist in %s\n"),
+			         Prog, line, fields[0], pw_dbname ());
 			errors++;
 			continue;
 		}
@@ -806,37 +912,42 @@ int main (int argc, char **argv)
 			errors++;
 			continue;
 		}
-		if (fields[4][0]) {
+		if ('\0' != fields[4][0]) {
 			newpw.pw_gecos = fields[4];
 		}
 
-		if (fields[5][0]) {
+		if ('\0' != fields[5][0]) {
 			newpw.pw_dir = fields[5];
 		}
 
-		if (fields[6][0]) {
+		if ('\0' != fields[6][0]) {
 			newpw.pw_shell = fields[6];
 		}
 
-		if (newpw.pw_dir[0] && access (newpw.pw_dir, F_OK)) {
-			if (mkdir (newpw.pw_dir,
-			           0777 & ~getdef_num ("UMASK",
-			                               GETDEF_DEFAULT_UMASK))) {
+		if (   ('\0' != newpw.pw_dir[0])
+		    && (access (newpw.pw_dir, F_OK) != 0)) {
+/* FIXME: should check for directory */
+			mode_t msk = 0777 & ~getdef_num ("UMASK",
+			                                 GETDEF_DEFAULT_UMASK);
+			if (mkdir (newpw.pw_dir, msk) != 0) {
 				fprintf (stderr,
-				         _("%s: line %d: mkdir failed\n"), Prog,
-				         line);
-			} else if (chown
-				   (newpw.pw_dir, newpw.pw_uid, newpw.pw_gid)) {
+				         _("%s: line %d: mkdir %s failed: %s\n"),
+				         Prog, line, newpw.pw_dir,
+				         strerror (errno));
+			} else if (chown (newpw.pw_dir,
+			                  newpw.pw_uid,
+			                  newpw.pw_gid) != 0) {
 				fprintf (stderr,
-				         _("%s: line %d: chown failed\n"), Prog,
-				         line);
+				         _("%s: line %d: chown %s failed: %s\n"),
+				         Prog, line, newpw.pw_dir,
+				         strerror (errno));
 			}
 		}
 
 		/*
 		 * Update the password entry with the new changes made.
 		 */
-		if (!pw_update (&newpw)) {
+		if (pw_update (&newpw) == 0) {
 			fprintf (stderr,
 			         _("%s: line %d: can't update entry\n"),
 			         Prog, line);
@@ -852,14 +963,9 @@ int main (int argc, char **argv)
 	 * changes to be written out all at once, and then unlocked
 	 * afterwards.
 	 */
-	if (errors) {
+	if (0 != errors) {
 		fprintf (stderr,
 		         _("%s: error detected, changes ignored\n"), Prog);
-		(void) gr_unlock ();
-		if (is_shadow) {
-			spw_unlock ();
-		}
-		(void) pw_unlock ();
 		fail_exit (1);
 	}
 
@@ -868,10 +974,6 @@ int main (int argc, char **argv)
 	nscd_flush_cache ("passwd");
 	nscd_flush_cache ("group");
 
-#ifdef USE_PAM
-	pam_end (pamh, PAM_SUCCESS);
-#endif				/* USE_PAM */
-
 	return 0;
 }
 
