@@ -32,7 +32,7 @@
 
 #include <config.h>
 
-#ident "$Id: userdel.c 2007 2008-05-24 15:35:15Z nekral-guest $"
+#ident "$Id: userdel.c 2649 2009-04-11 16:52:45Z nekral-guest $"
 
 #include <errno.h>
 #include <fcntl.h>
@@ -43,9 +43,11 @@
 #include <stdio.h>
 #include <sys/stat.h>
 #include <sys/stat.h>
+#ifdef ACCT_TOOLS_SETUID
 #ifdef USE_PAM
 #include "pam_defs.h"
 #endif				/* USE_PAM */
+#endif				/* ACCT_TOOLS_SETUID */
 #include "defines.h"
 #include "getdef.h"
 #include "groupio.h"
@@ -68,18 +70,28 @@
 #define E_USER_BUSY	8	/* user currently logged in */
 #define E_GRP_UPDATE	10	/* can't update group file */
 #define E_HOMEDIR	12	/* can't remove home directory */
+
+/*
+ * Global variables
+ */
+char *Prog;
+
 static char *user_name;
 static uid_t user_id;
 static char *user_home;
 
-static char *Prog;
-static int fflg = 0, rflg = 0;
+static bool fflg = false;
+static bool rflg = false;
 
-static int is_shadow_pwd;
+static bool is_shadow_pwd;
 
 #ifdef SHADOWGRP
-static int is_shadow_grp;
+static bool is_shadow_grp;
+static bool sgr_locked = false;
 #endif
+static bool pw_locked  = false;
+static bool gr_locked   = false;
+static bool spw_locked  = false;
 
 /* local function prototypes */
 static void usage (void);
@@ -92,7 +104,7 @@ static void user_busy (const char *, uid_t);
 static void user_cancel (const char *);
 
 #ifdef EXTRA_CHECK_HOME_DIR
-static int path_prefix (const char *, const char *);
+static bool path_prefix (const char *, const char *);
 #endif
 static int is_owner (uid_t, const char *);
 static void remove_mailbox (void);
@@ -130,7 +142,7 @@ static void update_groups (void)
 	struct passwd *pwd;
 
 #ifdef	SHADOWGRP
-	int deleted_user_group = 0;
+	bool deleted_user_group = false;
 	const struct sgrp *sgrp;
 	struct sgrp *nsgrp;
 #endif				/* SHADOWGRP */
@@ -139,30 +151,32 @@ static void update_groups (void)
 	 * Scan through the entire group file looking for the groups that
 	 * the user is a member of.
 	 */
-	for (gr_rewind (), grp = gr_next (); grp; grp = gr_next ()) {
+	for (gr_rewind (), grp = gr_next (); NULL != grp; grp = gr_next ()) {
 
 		/*
 		 * See if the user specified this group as one of their
 		 * concurrent groups.
 		 */
-		if (!is_on_list (grp->gr_mem, user_name))
+		if (!is_on_list (grp->gr_mem, user_name)) {
 			continue;
+		}
 
 		/* 
 		 * Delete the username from the list of group members and
 		 * update the group entry to reflect the change.
 		 */
 		ngrp = __gr_dup (grp);
-		if (!ngrp) {
+		if (NULL == ngrp) {
 			fprintf (stderr,
-				 _("%s: Out of memory. Cannot update the group database.\n"),
-				 Prog);
+			         _("%s: Out of memory. Cannot update %s.\n"),
+			         Prog, gr_dbname ());
 			exit (13);	/* XXX */
 		}
 		ngrp->gr_mem = del_list (ngrp->gr_mem, user_name);
-		if (!gr_update (ngrp)) {
+		if (gr_update (ngrp) == 0) {
 			fprintf (stderr,
-				 _("%s: error updating group entry\n"), Prog);
+			         _("%s: failed to prepare the new %s entry '%s'\n"),
+			         Prog, gr_dbname (), ngrp->gr_name);
 			exit (E_GRP_UPDATE);
 		}
 
@@ -170,11 +184,12 @@ static void update_groups (void)
 		 * Update the DBM group file with the new entry as well.
 		 */
 #ifdef WITH_AUDIT
-		audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
-			      "deleting user from group", user_name, user_id,
-			      0);
+		audit_logger (AUDIT_DEL_USER, Prog,
+		              "deleting user from group",
+		              user_name, (unsigned int) user_id,
+		              SHADOW_AUDIT_SUCCESS);
 #endif
-		SYSLOG ((LOG_INFO, "delete `%s' from group `%s'\n",
+		SYSLOG ((LOG_INFO, "delete '%s' from group '%s'\n",
 			 user_name, ngrp->gr_name));
 	}
 
@@ -184,8 +199,9 @@ static void update_groups (void)
 	 * user name, with no members, we delete it.
 	 */
 	grp = xgetgrnam (user_name);
-	if (grp && getdef_bool ("USERGROUPS_ENAB")
-	    && (grp->gr_mem[0] == NULL)) {
+	if (   (NULL != grp)
+	    && getdef_bool ("USERGROUPS_ENAB")
+	    && (NULL == grp->gr_mem[0])) {
 
 		pwd = NULL;
 		if (!fflg) {
@@ -194,51 +210,61 @@ static void update_groups (void)
 			 * used as a primary group.
 			 */
 			setpwent ();
-			while ((pwd = getpwent ())) {
-				if (strcmp (pwd->pw_name, user_name) == 0)
+			while ((pwd = getpwent ()) != NULL) {
+				if (strcmp (pwd->pw_name, user_name) == 0) {
 					continue;
+				}
 				if (pwd->pw_gid == grp->gr_gid) {
 					fprintf (stderr,
-						 _
-						 ("%s: Cannot remove group %s which is a primary group for another user.\n"),
-						 Prog, grp->gr_name);
+					         _("%s: group %s is the primary group of another user and is not removed.\n"),
+					         Prog, grp->gr_name);
 					break;
 				}
 			}
 			endpwent ();
 		}
 
-		if (pwd == NULL) {
+		if (NULL == pwd) {
 			/*
 			 * We can remove this group, it is not the primary
 			 * group of any remaining user.
 			 */
-			gr_remove (grp->gr_name);
+			if (gr_remove (grp->gr_name) == 0) {
+				fprintf (stderr,
+				         _("%s: cannot remove entry '%s' from %s\n"),
+				         Prog, grp->gr_name, gr_dbname ());
+				fail_exit (E_GRP_UPDATE);
+			}
 
 #ifdef SHADOWGRP
-			deleted_user_group = 1;
+			deleted_user_group = true;
 #endif
 
 #ifdef WITH_AUDIT
-			audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
-				      "deleting group", user_name, user_id, 0);
+			audit_logger (AUDIT_DEL_GROUP, Prog,
+			              "deleting group",
+			              grp->gr_name, AUDIT_NO_ID,
+			              SHADOW_AUDIT_SUCCESS);
 #endif
 			SYSLOG ((LOG_INFO,
-				 "removed group `%s' owned by `%s'\n",
+				 "removed group '%s' owned by '%s'\n",
 				 grp->gr_name, user_name));
 		}
 	}
 #ifdef	SHADOWGRP
-	if (!is_shadow_grp)
+	if (!is_shadow_grp) {
 		return;
+	}
 
 	/*
 	 * Scan through the entire shadow group file looking for the groups
 	 * that the user is a member of. Both the administrative list and
 	 * the ordinary membership list is checked.
 	 */
-	for (sgr_rewind (), sgrp = sgr_next (); sgrp; sgrp = sgr_next ()) {
-		int was_member, was_admin;
+	for (sgr_rewind (), sgrp = sgr_next ();
+	     NULL != sgrp;
+	     sgrp = sgr_next ()) {
+		bool was_member, was_admin;
 
 		/*
 		 * See if the user specified this group as one of their
@@ -247,39 +273,51 @@ static void update_groups (void)
 		was_member = is_on_list (sgrp->sg_mem, user_name);
 		was_admin = is_on_list (sgrp->sg_adm, user_name);
 
-		if (!was_member && !was_admin)
+		if (!was_member && !was_admin) {
 			continue;
+		}
 
 		nsgrp = __sgr_dup (sgrp);
-		if (!nsgrp) {
+		if (NULL == nsgrp) {
 			fprintf (stderr,
-				 _("%s: Out of memory. Cannot update the shadow group database.\n"),
-				 Prog);
+			         _("%s: Out of memory. Cannot update %s.\n"),
+			         Prog, sgr_dbname ());
 			exit (13);	/* XXX */
 		}
 
-		if (was_member)
+		if (was_member) {
 			nsgrp->sg_mem = del_list (nsgrp->sg_mem, user_name);
+		}
 
-		if (was_admin)
+		if (was_admin) {
 			nsgrp->sg_adm = del_list (nsgrp->sg_adm, user_name);
+		}
 
-		if (!sgr_update (nsgrp)) {
+		if (sgr_update (nsgrp) == 0) {
 			fprintf (stderr,
-				 _("%s: error updating shadow group entry\n"), Prog);
+			         _("%s: failed to prepare the new %s entry '%s'\n"),
+			         Prog, sgr_dbname (), nsgrp->sg_name);
 			exit (E_GRP_UPDATE);
 		}
 #ifdef WITH_AUDIT
-		audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
-			      "deleting user from shadow group", user_name,
-			      user_id, 0);
+		audit_logger (AUDIT_DEL_USER, Prog,
+		              "deleting user from shadow group",
+		              user_name, (unsigned int) user_id,
+		              SHADOW_AUDIT_SUCCESS);
 #endif
-		SYSLOG ((LOG_INFO, "delete `%s' from shadow group `%s'\n",
+		SYSLOG ((LOG_INFO, "delete '%s' from shadow group '%s'\n",
 			 user_name, nsgrp->sg_name));
 	}
 
-	if (deleted_user_group)
-		sgr_remove (user_name);
+	if (   deleted_user_group
+	    && (sgr_locate (user_name) != NULL)) {
+		if (sgr_remove (user_name) == 0) {
+			fprintf (stderr,
+			         _("%s: cannot remove entry '%s' from %s\n"),
+			         Prog, user_name, sgr_dbname ());
+			fail_exit (E_GRP_UPDATE);
+		}
+	}
 #endif				/* SHADOWGRP */
 }
 
@@ -291,26 +329,62 @@ static void update_groups (void)
  */
 static void close_files (void)
 {
-	if (!pw_close ())
-		fprintf (stderr, _("%s: cannot rewrite password file\n"), Prog);
-	if (is_shadow_pwd && !spw_close ())
-		fprintf (stderr,
-			 _("%s: cannot rewrite shadow password file\n"), Prog);
-	if (!gr_close ())
-		fprintf (stderr, _("%s: cannot rewrite group file\n"), Prog);
+	if (pw_close () == 0) {
+		fprintf (stderr, _("%s: failure while writing changes to %s\n"), Prog, pw_dbname ());
+		SYSLOG ((LOG_ERR, "failure while writing changes to %s", pw_dbname ()));
+		fail_exit (E_PW_UPDATE);
+	}
+	if (pw_unlock () == 0) {
+		fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, pw_dbname ());
+		SYSLOG ((LOG_ERR, "failed to unlock %s", pw_dbname ()));
+		/* continue */
+	}
+	pw_locked = false;
+
+	if (is_shadow_pwd) {
+		if (spw_close () == 0) {
+			fprintf (stderr,
+			         _("%s: failure while writing changes to %s\n"), Prog, spw_dbname ());
+			SYSLOG ((LOG_ERR, "failure while writing changes to %s", spw_dbname ()));
+			fail_exit (E_PW_UPDATE);
+		}
+		if (spw_unlock () == 0) {
+			fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, spw_dbname ());
+			SYSLOG ((LOG_ERR, "failed to unlock %s", spw_dbname ()));
+			/* continue */
+		}
+		spw_locked = false;
+	}
+
+	if (gr_close () == 0) {
+		fprintf (stderr, _("%s: failure while writing changes to %s\n"), Prog, gr_dbname ());
+		SYSLOG ((LOG_ERR, "failure while writing changes to %s", gr_dbname ()));
+		fail_exit (E_GRP_UPDATE);
+	}
+	if (gr_unlock () == 0) {
+		fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, gr_dbname ());
+		SYSLOG ((LOG_ERR, "failed to unlock %s", gr_dbname ()));
+		/* continue */
+	}
+	gr_locked = false;
 
-	(void) gr_unlock ();
 #ifdef	SHADOWGRP
-	if (is_shadow_grp && !sgr_close ())
-		fprintf (stderr,
-			 _("%s: cannot rewrite shadow group file\n"), Prog);
+	if (is_shadow_grp) {
+		if (sgr_close () == 0) {
+			fprintf (stderr,
+			         _("%s: failure while writing changes to %s\n"), Prog, sgr_dbname ());
+			SYSLOG ((LOG_ERR, "failure while writing changes to %s", sgr_dbname ()));
+			fail_exit (E_GRP_UPDATE);
+		}
 
-	if (is_shadow_grp)
-		(void) sgr_unlock ();
+		if (sgr_unlock () == 0) {
+			fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, sgr_dbname ());
+			SYSLOG ((LOG_ERR, "failed to unlock %s", sgr_dbname ()));
+			/* continue */
+		}
+		sgr_locked = false;
+	}
 #endif
-	if (is_shadow_pwd)
-		(void) spw_unlock ();
-	(void) pw_unlock ();
 }
 
 /*
@@ -318,18 +392,44 @@ static void close_files (void)
  */
 static void fail_exit (int code)
 {
-	(void) pw_unlock ();
-	(void) gr_unlock ();
-	if (is_shadow_pwd)
-		spw_unlock ();
+	if (pw_locked) {
+		if (pw_unlock () == 0) {
+			fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, pw_dbname ());
+			SYSLOG ((LOG_ERR, "failed to unlock %s", pw_dbname ()));
+			/* continue */
+		}
+	}
+	if (gr_locked) {
+		if (gr_unlock () == 0) {
+			fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, gr_dbname ());
+			SYSLOG ((LOG_ERR, "failed to unlock %s", gr_dbname ()));
+			/* continue */
+		}
+	}
+	if (spw_locked) {
+		if (spw_unlock () == 0) {
+			fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, spw_dbname ());
+			SYSLOG ((LOG_ERR, "failed to unlock %s", spw_dbname ()));
+			/* continue */
+		}
+	}
 #ifdef	SHADOWGRP
-	if (is_shadow_grp)
-		sgr_unlock ();
+	if (sgr_locked) {
+		if (sgr_unlock () == 0) {
+			fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, sgr_dbname ());
+			SYSLOG ((LOG_ERR, "failed to unlock %s", sgr_dbname ()));
+			/* continue */
+		}
+	}
 #endif
+
 #ifdef WITH_AUDIT
-	audit_logger (AUDIT_USER_CHAUTHTOK, Prog, "deleting user", user_name,
-		      user_id, 0);
+	audit_logger (AUDIT_DEL_USER, Prog,
+	              "deleting user",
+	              user_name, (unsigned int) user_id,
+	              SHADOW_AUDIT_FAILURE);
 #endif
+
 	exit (code);
 }
 
@@ -341,78 +441,106 @@ static void fail_exit (int code)
 
 static void open_files (void)
 {
-	if (!pw_lock ()) {
-		fprintf (stderr, _("%s: unable to lock password file\n"), Prog);
-#ifdef WITH_AUDIT
-		audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
-			      "locking password file", user_name, user_id, 0);
-#endif
-		exit (E_PW_UPDATE);
-	}
-	if (!pw_open (O_RDWR)) {
-		fprintf (stderr, _("%s: unable to open password file\n"), Prog);
-#ifdef WITH_AUDIT
-		audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
-			      "opening password file", user_name, user_id, 0);
-#endif
-		fail_exit (E_PW_UPDATE);
-	}
-	if (is_shadow_pwd && !spw_lock ()) {
+	if (pw_lock () == 0) {
 		fprintf (stderr,
-			 _("%s: cannot lock shadow password file\n"), Prog);
+		         _("%s: cannot lock %s; try again later.\n"),
+		         Prog, pw_dbname ());
 #ifdef WITH_AUDIT
-		audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
-			      "locking shadow password file", user_name,
-			      user_id, 0);
+		audit_logger (AUDIT_DEL_USER, Prog,
+		              "locking password file",
+		              user_name, (unsigned int) user_id,
+		              SHADOW_AUDIT_FAILURE);
 #endif
 		fail_exit (E_PW_UPDATE);
 	}
-	if (is_shadow_pwd && !spw_open (O_RDWR)) {
+	pw_locked = true;
+	if (pw_open (O_RDWR) == 0) {
 		fprintf (stderr,
-			 _("%s: cannot open shadow password file\n"), Prog);
+		         _("%s: cannot open %s\n"), Prog, pw_dbname ());
 #ifdef WITH_AUDIT
-		audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
-			      "opening shadow password file", user_name,
-			      user_id, 0);
+		audit_logger (AUDIT_DEL_USER, Prog,
+		              "opening password file",
+		              user_name, (unsigned int) user_id,
+		              SHADOW_AUDIT_FAILURE);
 #endif
 		fail_exit (E_PW_UPDATE);
 	}
-	if (!gr_lock ()) {
-		fprintf (stderr, _("%s: unable to lock group file\n"), Prog);
+	if (is_shadow_pwd) {
+		if (spw_lock () == 0) {
+			fprintf (stderr,
+			         _("%s: cannot lock %s; try again later.\n"),
+			         Prog, spw_dbname ());
 #ifdef WITH_AUDIT
-		audit_logger (AUDIT_USER_CHAUTHTOK, Prog, "locking group file",
-			      user_name, user_id, 0);
+			audit_logger (AUDIT_DEL_USER, Prog,
+			              "locking shadow password file",
+			              user_name, (unsigned int) user_id,
+			              SHADOW_AUDIT_FAILURE);
+#endif
+			fail_exit (E_PW_UPDATE);
+		}
+		spw_locked = true;
+		if (spw_open (O_RDWR) == 0) {
+			fprintf (stderr,
+			         _("%s: cannot open %s\n"),
+			         Prog, spw_dbname ());
+#ifdef WITH_AUDIT
+			audit_logger (AUDIT_DEL_USER, Prog,
+			              "opening shadow password file",
+			              user_name, (unsigned int) user_id,
+			              SHADOW_AUDIT_FAILURE);
+#endif
+			fail_exit (E_PW_UPDATE);
+		}
+	}
+	if (gr_lock () == 0) {
+		fprintf (stderr,
+		         _("%s: cannot lock %s; try again later.\n"),
+		         Prog, gr_dbname ());
+#ifdef WITH_AUDIT
+		audit_logger (AUDIT_DEL_USER, Prog,
+		              "locking group file",
+		              user_name, (unsigned int) user_id,
+		              SHADOW_AUDIT_FAILURE);
 #endif
 		fail_exit (E_GRP_UPDATE);
 	}
-	if (!gr_open (O_RDWR)) {
-		fprintf (stderr, _("%s: cannot open group file\n"), Prog);
+	gr_locked = true;
+	if (gr_open (O_RDWR) == 0) {
+		fprintf (stderr, _("%s: cannot open %s\n"), Prog, gr_dbname ());
 #ifdef WITH_AUDIT
-		audit_logger (AUDIT_USER_CHAUTHTOK, Prog, "opening group file",
-			      user_name, user_id, 0);
+		audit_logger (AUDIT_DEL_USER, Prog,
+		              "opening group file",
+		              user_name, (unsigned int) user_id,
+		              SHADOW_AUDIT_FAILURE);
 #endif
 		fail_exit (E_GRP_UPDATE);
 	}
 #ifdef	SHADOWGRP
-	if (is_shadow_grp && !sgr_lock ()) {
-		fprintf (stderr,
-			 _("%s: unable to lock shadow group file\n"), Prog);
+	if (is_shadow_grp) {
+		if (sgr_lock () == 0) {
+			fprintf (stderr,
+			         _("%s: cannot lock %s; try again later.\n"),
+			         Prog, sgr_dbname ());
 #ifdef WITH_AUDIT
-		audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
-			      "locking shadow group file", user_name, user_id,
-			      0);
+			audit_logger (AUDIT_DEL_USER, Prog,
+			              "locking shadow group file",
+			              user_name, (unsigned int) user_id,
+			              SHADOW_AUDIT_FAILURE);
 #endif
-		fail_exit (E_GRP_UPDATE);
-	}
-	if (is_shadow_grp && !sgr_open (O_RDWR)) {
-		fprintf (stderr, _("%s: cannot open shadow group file\n"),
-			 Prog);
+			fail_exit (E_GRP_UPDATE);
+		}
+		sgr_locked= true;
+		if (sgr_open (O_RDWR) == 0) {
+			fprintf (stderr, _("%s: cannot open %s\n"),
+			         Prog, sgr_dbname ());
 #ifdef WITH_AUDIT
-		audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
-			      "opening shadow group file", user_name, user_id,
-			      0);
+			audit_logger (AUDIT_DEL_USER, Prog,
+			              "opening shadow group file",
+			              user_name, (unsigned int) user_id,
+			              SHADOW_AUDIT_FAILURE);
 #endif
-		fail_exit (E_GRP_UPDATE);
+			fail_exit (E_GRP_UPDATE);
+		}
 	}
 #endif
 }
@@ -425,21 +553,27 @@ static void open_files (void)
  */
 static void update_user (void)
 {
-	if (!pw_remove (user_name)) {
+	if (pw_remove (user_name) == 0) {
 		fprintf (stderr,
-			 _("%s: error deleting password entry\n"), Prog);
+		         _("%s: cannot remove entry '%s' from %s\n"),
+		         Prog, user_name, pw_dbname ());
 		fail_exit (E_PW_UPDATE);
 	}
-	if (is_shadow_pwd && !spw_remove (user_name)) {
+	if (   is_shadow_pwd
+	    && (spw_locate (user_name) != NULL)
+	    && (spw_remove (user_name) == 0)) {
 		fprintf (stderr,
-			 _("%s: error deleting shadow password entry\n"), Prog);
+		         _("%s: cannot remove entry '%s' from %s\n"),
+		         Prog, user_name, spw_dbname ());
 		fail_exit (E_PW_UPDATE);
 	}
 #ifdef WITH_AUDIT
-	audit_logger (AUDIT_USER_CHAUTHTOK, Prog, "deleting user entries",
-		      user_name, user_id, 1);
+	audit_logger (AUDIT_DEL_USER, Prog,
+	              "deleting user entries",
+	              user_name, (unsigned int) user_id,
+	              SHADOW_AUDIT_SUCCESS);
 #endif
-	SYSLOG ((LOG_INFO, "delete user `%s'\n", user_name));
+	SYSLOG ((LOG_INFO, "delete user '%s'\n", user_name));
 }
 
 /*
@@ -460,24 +594,28 @@ static void user_busy (const char *name, uid_t uid)
 	struct utmpx *utent;
 
 	setutxent ();
-	while ((utent = getutxent ())) {
+	while ((utent = getutxent ()) != NULL)
 #else
 	struct utmp *utent;
 
 	setutent ();
-	while ((utent = getutent ())) {
+	while ((utent = getutent ()) != NULL)
 #endif
+	{
 		if (utent->ut_type != USER_PROCESS)
 			continue;
 
-		if (strncmp (utent->ut_user, name, sizeof utent->ut_user))
+		if (strncmp (utent->ut_user, name, sizeof utent->ut_user) != 0) {
 			continue;
+		}
 		fprintf (stderr,
 			 _("%s: user %s is currently logged in\n"), Prog, name);
 		if (!fflg) {
 #ifdef WITH_AUDIT
-			audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
-				      "deleting user logged in", name, -1, 0);
+			audit_logger (AUDIT_DEL_USER, Prog,
+			              "deleting user logged in",
+			              name, AUDIT_NO_ID,
+			              SHADOW_AUDIT_FAILURE);
 #endif
 			exit (E_USER_BUSY);
 		}
@@ -524,30 +662,33 @@ exit 0
 static void user_cancel (const char *user)
 {
 	char *cmd;
-	int pid, wpid;
+	pid_t pid, wpid;
 	int status;
 
-	if (!(cmd = getdef_str ("USERDEL_CMD")))
+	cmd = getdef_str ("USERDEL_CMD");
+	if (NULL == cmd) {
 		return;
+	}
 	pid = fork ();
 	if (pid == 0) {
 		execl (cmd, cmd, user, (char *) 0);
 		perror (cmd);
 		_exit (errno == ENOENT ? E_CMD_NOTFOUND : E_CMD_NOEXEC);
-	} else if (pid == -1) {
+	} else if ((pid_t)-1 == pid) {
 		perror ("fork");
 		return;
 	}
 	do {
 		wpid = wait (&status);
-	} while (wpid != pid && wpid != -1);
+	} while ((wpid != pid) && ((pid_t)-1 != wpid));
 }
 
 #ifdef EXTRA_CHECK_HOME_DIR
-static int path_prefix (const char *s1, const char *s2)
+static bool path_prefix (const char *s1, const char *s2)
 {
-	return (strncmp (s2, s1, strlen (s1)) == 0 &&
-		(s2[strlen (s1)] == '\0' || s2[strlen (s1)] == '/'));
+	return (   (strncmp (s2, s1, strlen (s1)) == 0)
+	        && (   ('\0' == s2[strlen (s1)])
+	            || ('/'  == s2[strlen (s1)])));
 }
 #endif
 
@@ -555,8 +696,9 @@ static int is_owner (uid_t uid, const char *path)
 {
 	struct stat st;
 
-	if (stat (path, &st))
+	if (stat (path, &st) != 0) {
 		return -1;
+	}
 	return (st.st_uid == uid);
 }
 
@@ -568,41 +710,71 @@ static void remove_mailbox (void)
 
 	maildir = getdef_str ("MAIL_DIR");
 #ifdef MAIL_SPOOL_DIR
-	if (!maildir && !getdef_str ("MAIL_FILE"))
+	if ((NULL == maildir) && (getdef_str ("MAIL_FILE") == NULL)) {
 		maildir = MAIL_SPOOL_DIR;
+	}
 #endif
-	if (!maildir)
+	if (NULL == maildir) {
 		return;
+	}
 	snprintf (mailfile, sizeof mailfile, "%s/%s", maildir, user_name);
 	if (fflg) {
-		unlink (mailfile);	/* always remove, ignore errors */
+		if (unlink (mailfile) != 0) {
+			fprintf (stderr, _("%s: warning: can't remove %s: %s"), Prog, mailfile, strerror (errno));
+			SYSLOG ((LOG_ERR, "Cannot remove %s: %s", mailfile, strerror (errno)));
 #ifdef WITH_AUDIT
-		audit_logger (AUDIT_USER_CHAUTHTOK, Prog, "deleting mail file",
-			      user_name, user_id, 1);
+			audit_logger (AUDIT_DEL_USER, Prog,
+			              "deleting mail file",
+			              user_name, (unsigned int) user_id,
+			              SHADOW_AUDIT_FAILURE);
+#endif
+			/* continue */
+		}
+#ifdef WITH_AUDIT
+		else
+		{
+			audit_logger (AUDIT_DEL_USER, Prog,
+			              "deleting mail file",
+			              user_name, (unsigned int) user_id,
+			              SHADOW_AUDIT_SUCCESS);
+		}
 #endif
 		return;
 	}
 	i = is_owner (user_id, mailfile);
 	if (i == 0) {
 		fprintf (stderr,
-			 _
-			 ("%s: %s not owned by %s, not removing\n"),
-			 Prog, mailfile, user_name);
+		         _("%s: %s not owned by %s, not removing\n"),
+		         Prog, mailfile, user_name);
+		SYSLOG ((LOG_ERR, "%s not owned by %s, not removed", mailfile, strerror (errno)));
 #ifdef WITH_AUDIT
-		audit_logger (AUDIT_USER_CHAUTHTOK, Prog, "deleting mail file",
-			      user_name, user_id, 0);
+		audit_logger (AUDIT_DEL_USER, Prog,
+		              "deleting mail file",
+		              user_name, (unsigned int) user_id,
+		              SHADOW_AUDIT_FAILURE);
 #endif
 		return;
-	} else if (i == -1)
+	} else if (i == -1) {
 		return;		/* mailbox doesn't exist */
-	if (unlink (mailfile)) {
-		fprintf (stderr, _("%s: warning: can't remove "), Prog);
-		perror (mailfile);
+	}
+	if (unlink (mailfile) != 0) {
+		fprintf (stderr, _("%s: warning: can't remove %s: %s"), Prog, mailfile, strerror (errno));
+		SYSLOG ((LOG_ERR, "Cannot remove %s: %s", mailfile, strerror (errno)));
+#ifdef WITH_AUDIT
+		audit_logger (AUDIT_DEL_USER, Prog,
+		              "deleting mail file",
+		              user_name, (unsigned int) user_id,
+		              SHADOW_AUDIT_FAILURE);
+#endif
+		/* continue */
 	}
 #ifdef WITH_AUDIT
-	else {
-		audit_logger (AUDIT_USER_CHAUTHTOK, Prog, "deleting mail file",
-			      user_name, user_id, 1);
+	else
+	{
+		audit_logger (AUDIT_DEL_USER, Prog,
+		              "deleting mail file",
+		              user_name, (unsigned int) user_id,
+		              SHADOW_AUDIT_SUCCESS);
 	}
 #endif
 }
@@ -612,24 +784,37 @@ static void remove_mailbox (void)
  */
 int main (int argc, char **argv)
 {
-	int errors = 0;
+	int errors = 0; /* Error in the removal of the home directory */
 
+#ifdef ACCT_TOOLS_SETUID
 #ifdef USE_PAM
 	pam_handle_t *pamh = NULL;
 	int retval;
-#endif
+#endif				/* USE_PAM */
+#endif				/* ACCT_TOOLS_SETUID */
 
 #ifdef WITH_AUDIT
 	audit_help_open ();
 #endif
 
+#ifdef WITH_SELINUX
+	if (is_selinux_enabled () > 0) {
+		const char *args[5];
+		args[0] = "/usr/sbin/semanage";
+		args[1] = "login";
+		args[2] = "-d";
+		args[3] = user_name;
+		args[4] = NULL;
+		safe_system (args[0], args, NULL, 1);
+	}
+#endif
 	/*
 	 * Get my name so that I can use it to report errors.
 	 */
 	Prog = Basename (argv[0]);
-	setlocale (LC_ALL, "");
-	bindtextdomain (PACKAGE, LOCALEDIR);
-	textdomain (PACKAGE);
+	(void) setlocale (LC_ALL, "");
+	(void) bindtextdomain (PACKAGE, LOCALEDIR);
+	(void) textdomain (PACKAGE);
 
 	{
 		/*
@@ -642,15 +827,14 @@ int main (int argc, char **argv)
 			{"remove", no_argument, NULL, 'r'},
 			{NULL, 0, NULL, '\0'}
 		};
-		while ((c =
-			getopt_long (argc, argv, "fhr",
-				     long_options, NULL)) != -1) {
+		while ((c = getopt_long (argc, argv, "fhr",
+		                         long_options, NULL)) != -1) {
 			switch (c) {
 			case 'f':	/* force remove even if not owned by user */
-				fflg++;
+				fflg = true;
 				break;
 			case 'r':	/* remove home dir and mailbox */
-				rflg++;
+				rflg = true;
 				break;
 			default:
 				usage ();
@@ -658,44 +842,44 @@ int main (int argc, char **argv)
 		}
 	}
 
-	if (optind + 1 != argc)
+	if ((optind + 1) != argc) {
 		usage ();
+	}
 
 	OPENLOG ("userdel");
 
+#ifdef ACCT_TOOLS_SETUID
 #ifdef USE_PAM
-	retval = PAM_SUCCESS;
-
 	{
 		struct passwd *pampw;
 		pampw = getpwuid (getuid ()); /* local, no need for xgetpwuid */
 		if (pampw == NULL) {
-			retval = PAM_USER_UNKNOWN;
+			fprintf (stderr,
+			         _("%s: Cannot determine your user name.\n"),
+			         Prog);
+			exit (E_PW_UPDATE);
 		}
 
-		if (retval == PAM_SUCCESS) {
-			retval = pam_start ("userdel", pampw->pw_name,
-					    &conv, &pamh);
-		}
+		retval = pam_start ("userdel", pampw->pw_name, &conv, &pamh);
 	}
 
-	if (retval == PAM_SUCCESS) {
+	if (PAM_SUCCESS == retval) {
 		retval = pam_authenticate (pamh, 0);
-		if (retval != PAM_SUCCESS)
-			pam_end (pamh, retval);
 	}
 
-	if (retval == PAM_SUCCESS) {
+	if (PAM_SUCCESS == retval) {
 		retval = pam_acct_mgmt (pamh, 0);
-		if (retval != PAM_SUCCESS)
-			pam_end (pamh, retval);
 	}
 
-	if (retval != PAM_SUCCESS) {
+	if (NULL != pamh) {
+		(void) pam_end (pamh, retval);
+	}
+	if (PAM_SUCCESS != retval) {
 		fprintf (stderr, _("%s: PAM authentication failed\n"), Prog);
 		exit (E_PW_UPDATE);
 	}
 #endif				/* USE_PAM */
+#endif				/* ACCT_TOOLS_SETUID */
 
 	is_shadow_pwd = spw_file_present ();
 #ifdef SHADOWGRP
@@ -708,13 +892,15 @@ int main (int argc, char **argv)
 	user_name = argv[argc - 1];
 	{
 		struct passwd *pwd;
-		/* local, no need for xgetpwnam */
-		if (!(pwd = getpwnam (user_name))) {
-			fprintf (stderr, _("%s: user %s does not exist\n"),
+		pwd = getpwnam (user_name); /* local, no need for xgetpwnam */
+		if (NULL == pwd) {
+			fprintf (stderr, _("%s: user '%s' does not exist\n"),
 				 Prog, user_name);
 #ifdef WITH_AUDIT
-			audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
-				      "deleting user not found", user_name, -1, 0);
+			audit_logger (AUDIT_DEL_USER, Prog,
+			              "deleting user not found",
+			              user_name, AUDIT_NO_ID,
+			              SHADOW_AUDIT_FAILURE);
 #endif
 			exit (E_NOTFOUND);
 		}
@@ -732,7 +918,7 @@ int main (int argc, char **argv)
 
 		fprintf (stderr,
 			 _("%s: user %s is a NIS user\n"), Prog, user_name);
-		if (!yp_get_default_domain (&nis_domain)
+		if (   !yp_get_default_domain (&nis_domain)
 		    && !yp_master (nis_domain, "passwd.byname", &nis_master)) {
 			fprintf (stderr,
 				 _("%s: %s is the NIS master\n"),
@@ -754,14 +940,16 @@ int main (int argc, char **argv)
 	update_user ();
 	update_groups ();
 
-	if (rflg)
+	if (rflg) {
 		remove_mailbox ();
-	if (rflg && !fflg && !is_owner (user_id, user_home)) {
+	}
+	if (rflg && !fflg && (is_owner (user_id, user_home) == 0)) {
 		fprintf (stderr,
 			 _("%s: %s not owned by %s, not removing\n"),
 			 Prog, user_home, user_name);
 		rflg = 0;
 		errors++;
+		/* continue */
 	}
 
 #ifdef EXTRA_CHECK_HOME_DIR
@@ -777,15 +965,17 @@ int main (int argc, char **argv)
 		 */
 		setpwent ();
 		while ((pwd = getpwent ())) {
-			if (strcmp (pwd->pw_name, user_name) == 0)
+			if (strcmp (pwd->pw_name, user_name) == 0) {
 				continue;
+			}
 			if (path_prefix (user_home, pwd->pw_dir)) {
 				fprintf (stderr,
 					 _
 					 ("%s: not removing directory %s (would remove home of user %s)\n"),
 					 Prog, user_home, pwd->pw_name);
-				rflg = 0;
+				rflg = false;
 				errors++;
+				/* continue */
 				break;
 			}
 		}
@@ -798,18 +988,27 @@ int main (int argc, char **argv)
 			fprintf (stderr,
 				 _("%s: error removing directory %s\n"),
 				 Prog, user_home);
-#ifdef WITH_AUDIT
-			audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
-				      "deleting home directory", user_name,
-				      user_id, 1);
-#endif
 			errors++;
+			/* continue */
 		}
 #ifdef WITH_AUDIT
-		audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
-			      "deleting home directory", user_name, user_id, 1);
+		else
+		{
+			audit_logger (AUDIT_DEL_USER, Prog,
+			              "deleting home directory",
+			              user_name, (unsigned int) user_id,
+			              SHADOW_AUDIT_SUCCESS);
+		}
 #endif
 	}
+#ifdef WITH_AUDIT
+	if (0 != errors) {
+		audit_logger (AUDIT_DEL_USER, Prog,
+		              "deleting home directory",
+		              user_name, AUDIT_NO_ID,
+		              SHADOW_AUDIT_FAILURE);
+	}
+#endif
 
 	/*
 	 * Cancel any crontabs or at jobs. Have to do this before we remove
@@ -821,15 +1020,7 @@ int main (int argc, char **argv)
 	nscd_flush_cache ("passwd");
 	nscd_flush_cache ("group");
 
-#ifdef USE_PAM
-	if (retval == PAM_SUCCESS)
-		pam_end (pamh, PAM_SUCCESS);
-#endif				/* USE_PAM */
-#ifdef WITH_AUDIT
-	if (errors)
-		audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
-			      "deleting home directory", user_name, -1, 0);
-#endif
-	exit (errors ? E_HOMEDIR : E_SUCCESS);
+	exit ((0 != errors) ? E_HOMEDIR : E_SUCCESS);
 	/* NOT REACHED */
 }
+
